InterpreterPlugin subclass: #OSProcessPlugin
	instanceVariableNames: 'osprocessSandboxSecurity'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMConstruction-Plugins-OSProcessPlugin'!

!OSProcessPlugin commentStamp: '<historical>' prior: 0!
I provide pluggable primitive access to the operating system process in which the
Squeak VM is currently running. One or more of my subclasses may be translated
to produce platform specific plugins.!

OSProcessPlugin subclass: #UnixOSProcessPlugin
	instanceVariableNames: 'sigChldSemaIndex sigHandlers originalSigHandlers semaIndices vmThread'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMConstruction-Plugins-OSProcessPlugin'!

!UnixOSProcessPlugin commentStamp: '<historical>' prior: 0!
I provide pluggable primitive access to the Unix operating system process in which the
Squeak VM is currently running.!

UnixOSProcessPlugin subclass: #UnixOSProcessPluginInterpreterGetThisSession
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMConstruction-Plugins-OSProcessPlugin'!

!UnixOSProcessPluginInterpreterGetThisSession commentStamp: 'dtl 8/7/2005 22:40' prior: 0!
This is a UnixOSProcessPlugin that obtains the value of thisSession from the interpreter.!

UnixOSProcessPlugin subclass: #UnixOSProcessPluginNoThisSessionAvailable
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMConstruction-Plugins-OSProcessPlugin'!

!UnixOSProcessPluginNoThisSessionAvailable commentStamp: 'dtl 8/7/2005 22:43' prior: 0!
This is a UnixOSProcessPlugin which does not know how to obtain the value of thisSession.
Use this only if it is not possible to obtain the session identifier from the interpreter.

This class is obsolete and will be removed in a future release of OSPP.
!

OSProcessPlugin subclass: #Win32OSProcessPlugin
	instanceVariableNames: 'readCharBufferArray readCharCountArray readCharStatusArray readCharArraySize readCharThreadIdArray'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMConstruction-Plugins-OSProcessPlugin'!

!Win32OSProcessPlugin commentStamp: '<historical>' prior: 0!
I provide pluggable primitive access to the Windows operating system process in which the
Squeak VM is currently running.!

Win32OSProcessPlugin class
	instanceVariableNames: 'arrayValueOf'!
Win32OSProcessPlugin subclass: #Win32OSProcessPluginInterpreterGetThisSession
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMConstruction-Plugins-OSProcessPlugin'!

!Win32OSProcessPluginInterpreterGetThisSession commentStamp: 'dtl 8/7/2005 22:41' prior: 0!
This is a Win32OSProcessPlugin that obtains the value of thisSession from the interpreter.!

Win32OSProcessPlugin subclass: #Win32OSProcessPluginNoThisSessionAvailable
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMConstruction-Plugins-OSProcessPlugin'!

!Win32OSProcessPluginNoThisSessionAvailable commentStamp: 'dtl 8/7/2005 22:43' prior: 0!
This is a Win32OSProcessPlugin which does not know how to obtain the value of thisSession.
Use this only if it is not possible to obtain the session identifier from the interpreter.

This class is obsolete and will be removed in a future release of OSPP.
!


!OSProcessPlugin methodsFor: 'initialize-release' stamp: 'dtl 8/13/2003 19:25'!
initialiseModule

	self export: true.
	osprocessSandboxSecurity _ -1
! !

!OSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 6/13/2005 00:19'!
primitiveGetSession
	"Answer the unique session identifier for this Smalltalk instance running in this
	OS process. The C integer value is coerced into a Smalltalk ByteArray to preserve
	the full range of possible values."

	| sessionOop sessionByteArrayPointer thisSessionID |
	self export: true.
	self var: 'sessionByteArrayPointer' declareC: 'unsigned char *sessionByteArrayPointer'.
	self var: 'copyByteArrayPointer' declareC: 'unsigned char *copyByteArrayPointer'.
	self var: 'thisSessionID' type: 'SESSIONIDENTIFIERTYPE'.
	sessionOop _ interpreterProxy instantiateClass: interpreterProxy classByteArray indexableSize: self sizeOfInt.
	sessionByteArrayPointer _ interpreterProxy arrayValueOf: sessionOop.
	thisSessionID _ self getThisSessionIdentifier.
	thisSessionID isNil ifTrue:
		[^ interpreterProxy primitiveFail].	"The session ID is not permitted to be zero"
	self copyBytesFrom: (self cCode: '(unsigned char *)&thisSessionID' inSmalltalk: [thisSessionID])
		to: sessionByteArrayPointer
		length: self sizeOfInt.
	interpreterProxy pop: 1; push: sessionOop
! !

!OSProcessPlugin methodsFor: 'primitives - pipes' stamp: 'dtl 6/13/2005 00:21'!
primitiveCreatePipe
	"Create a pipe, and answer an array of two file handles for the pipe writer and reader.
	The readerIOStream and writerIOStream variables represent the low level pipe streams,
	which will be of type (FILE *) or HANDLE, depending on what the FilePlugin support
	code is using to represent file streams. FILEHANDLETYPE is defined in my subclasses
	in the #declareCVarsIn: class method."

	| reader writer readerPtr writerPtr readerIOStream writerIOStream
	readerIOStreamPtr writerIOStreamPtr thisSession arrayResult |
	self export: true.
	self var: 'readerIOStream' type: 'FILEHANDLETYPE'.
	self var: 'writerIOStream' type: 'FILEHANDLETYPE'.
	self var: 'readerIOStreamPtr' declareC: 'FILEHANDLETYPE *readerIOStreamPtr'.
	self var: 'writerIOStreamPtr' declareC: 'FILEHANDLETYPE *writerIOStreamPtr'.
	self var: 'readerPtr' declareC: 'SQFile *readerPtr'.
	self var: 'writerPtr' declareC: 'SQFile *writerPtr'.
	self var: 'arrayPtr' declareC: 'int *arrayPtr'.
	self var: 'thisSession' type: 'SESSIONIDENTIFIERTYPE'.
	thisSession _ self getThisSessionIdentifier.

	"Create the anonymous OS pipe"
	self cCode: 'readerIOStreamPtr = &readerIOStream'.
	self cCode: 'writerIOStreamPtr = &writerIOStream'.
	(self createPipeForReader: readerIOStreamPtr writer: writerIOStreamPtr)
		ifFalse: [^ interpreterProxy primitiveFail].

	"Map the writer end to a SQFile data structure"
	writer _ self newSQFileByteArray.
	writerPtr _ self fileValueOf: writer.
	self cCode: 'writerPtr->file = writerIOStream'.
	self cCode: 'writerPtr->sessionID = thisSession'.
	self cCode: 'writerPtr->writable = 1'.
	self cCode: 'writerPtr->lastOp = 0'.
	interpreterProxy pushRemappableOop: writer.

	"Map the reader end to a SQFile data structure"
	reader _ self newSQFileByteArray.
	readerPtr _ self fileValueOf: reader.
	self cCode: 'readerPtr->file = readerIOStream'.	
	self cCode: 'readerPtr->sessionID = thisSession'.
	self cCode: 'readerPtr->writable = 0'.
	self cCode: 'readerPtr->lastOp = 0'.
	interpreterProxy pushRemappableOop: reader.

	"Answer an array of two SQFile pipe ends"
	arrayResult _ interpreterProxy instantiateClass: interpreterProxy classArray indexableSize: 2.
	interpreterProxy stObject: arrayResult at: 1 put: interpreterProxy popRemappableOop.	"reader"
	interpreterProxy stObject: arrayResult at: 2 put: interpreterProxy popRemappableOop. "writer"
	interpreterProxy pop: 1; push: arrayResult
! !

!OSProcessPlugin methodsFor: 'primitives - pipes' stamp: 'dtl 6/13/2005 00:21'!
primitiveCreatePipeWithSessionIdentifier
	"Create a pipe, and answer an array of two file handles for the pipe writer and reader.
	The session identifier is passed as the parameter to this primitive. Use this variant
	if the session identifier is not available directly in the VM (as may be the case if
	it is not possible to link from this plugin to a variable elsewhere in the VM).
	The readerIOStream and writerIOStream variables represent the low level pipe streams,
	which will be of type (FILE *) or HANDLE, depending on what the FilePlugin support
	code is using to represent file streams. FILEHANDLETYPE is defined in my subclasses
	in the #declareCVarsIn: class method."

	| reader writer readerPtr writerPtr readerIOStream writerIOStream
	readerIOStreamPtr writerIOStreamPtr thisSession arrayResult |
	self export: true.
	self var: 'readerIOStream' type: 'FILEHANDLETYPE'.
	self var: 'writerIOStream' type: 'FILEHANDLETYPE'.
	self var: 'readerIOStreamPtr' declareC: 'FILEHANDLETYPE *readerIOStreamPtr'.
	self var: 'writerIOStreamPtr' declareC: 'FILEHANDLETYPE *writerIOStreamPtr'.
	self var: 'readerPtr' declareC: 'SQFile *readerPtr'.
	self var: 'writerPtr' declareC: 'SQFile *writerPtr'.
	self var: 'arrayPtr' declareC: 'int *arrayPtr'.
	self var: 'thisSession' type: 'SESSIONIDENTIFIERTYPE'.
	thisSession _ self sessionIdentifierFrom: (interpreterProxy stackObjectValue: 0).

	"Create the anonymous OS pipe"
	self cCode: 'readerIOStreamPtr = &readerIOStream'.
	self cCode: 'writerIOStreamPtr = &writerIOStream'.
	(self createPipeForReader: readerIOStreamPtr writer: writerIOStreamPtr)
		ifFalse: [^ interpreterProxy primitiveFail].

	"Map the writer end to a SQFile data structure"
	writer _ self newSQFileByteArray.
	writerPtr _ self fileValueOf: writer.
	self cCode: 'writerPtr->file = writerIOStream'.
	self cCode: 'writerPtr->sessionID = thisSession'.
	self cCode: 'writerPtr->writable = 1'.
	self cCode: 'writerPtr->lastOp = 0'.
	interpreterProxy pushRemappableOop: writer.

	"Map the reader end to a SQFile data structure"
	reader _ self newSQFileByteArray.
	readerPtr _ self fileValueOf: reader.
	self cCode: 'readerPtr->file = readerIOStream'.	
	self cCode: 'readerPtr->sessionID = thisSession'.
	self cCode: 'readerPtr->writable = 0'.
	self cCode: 'readerPtr->lastOp = 0'.
	interpreterProxy pushRemappableOop: reader.

	"Answer an array of two SQFile pipe ends"
	arrayResult _ interpreterProxy instantiateClass: interpreterProxy classArray indexableSize: 2.
	interpreterProxy stObject: arrayResult at: 1 put: interpreterProxy popRemappableOop.	"reader"
	interpreterProxy stObject: arrayResult at: 2 put: interpreterProxy popRemappableOop. "writer"
	interpreterProxy pop: 2; push: arrayResult
! !

!OSProcessPlugin methodsFor: 'primitives - machine specific' stamp: 'dtl 7/14/2003 06:32'!
primitiveSizeOfInt
	"Size in bytes of an integer, for this C compiler on this machine."

	self export: true.
	interpreterProxy pop: 1; pushInteger: self sizeOfInt! !

!OSProcessPlugin methodsFor: 'primitives - machine specific' stamp: 'dtl 7/14/2003 06:33'!
primitiveSizeOfPointer
	"Size in bytes of a void pointer, for this C compiler on this machine."

	self export: true.
	interpreterProxy pop: 1; pushInteger: self sizeOfPointer! !

!OSProcessPlugin methodsFor: 'primitives - version testing' stamp: 'dtl 2/21/2002 16:29'!
primitiveModuleName
	"Answer a string containing the module name string for this plugin."

	| s |
	self export: true.
	self var: 's' declareC: 'char *s'.
	self cCode: 's= (char *)moduleName' inSmalltalk: [s _ moduleName]. "Prevent compiler warning"
	interpreterProxy pop: 1; push: (self stringFromCString: s)
! !

!OSProcessPlugin methodsFor: 'primitives - version testing' stamp: 'dtl 3/1/2002 16:56'!
primitiveVersionString
	"Answer a string containing the version string for this plugin."

	self export: true.
	interpreterProxy pop: 1; push: (self stringFromCString: self versionString)
! !

!OSProcessPlugin methodsFor: 'version identification' stamp: 'dtl 11/5/2005 07:51'!
versionString
	"Answer a string containing the version string for this plugin. Handle MNU
	errors, which can occur if class InterpreterPlugin has been removed from
	the system.

	Important: When this method is changed, the class side method must also be
	changed to match."

	"4.0 supports 64bit code base"

	| version |
	self returnTypeC: 'char *'.
	self var: 'version' declareC: 'static char version[]= "4.0.1"'.
	^ self cCode: 'version' inSmalltalk: [self class versionString]
! !

!OSProcessPlugin methodsFor: 'private' stamp: 'dtl 2/21/2002 16:30'!
callocWrapper: count size: objectSize
	"Using malloc() and calloc() is something I would like to avoid, since it is
	likely to cause problems some time in the future if somebody redesigns
	object memory allocation. This wrapper just makes it easy to find senders
	of calloc() in my code. -dtl"

	self returnTypeC: 'void *'.
	^ self cCode: 'calloc(count, objectSize)'! !

!OSProcessPlugin methodsFor: 'private' stamp: 'dtl 2/21/2002 16:30'!
cStringFromString: aString
	"Answer a new null-terminated C string copied from aString. The C string
	is allocated from the C runtime heap. See transientCStringFromString for
	a version which allocates from object memory.
	Caution: This may invoke the garbage collector."

	| len sPtr cString |
	self returnTypeC: 'char *'.
	self var: 'sPtr' declareC: 'char *sPtr'.
	self var: 'cString' declareC: 'char *cString'.
	sPtr _ interpreterProxy arrayValueOf: aString.
	len _ interpreterProxy sizeOfSTArrayFromCPrimitive: sPtr.
	cString _ self callocWrapper: len + 1 size: 1.			"Space for a null terminated C string."
	self cCode: '(char *) strncpy (cString, sPtr, len)'.	"Copy the string."
	^ cString
! !

!OSProcessPlugin methodsFor: 'private' stamp: 'dtl 7/12/2005 07:05'!
fixPointersInArrayOfStrings: flattenedArrayOfStrings withOffsets: offsetArray count: count
	"Use the address offsets in offsetArray to fix up the pointers in cStringArray.
	The result is a C array of pointers to char, used for argv and env vectors."

	| ptr idx |
	self var: 'flattenedArrayOfStrings' declareC: 'char *flattenedArrayOfStrings'.
	self var: 'offsetArray' declareC: 'sqInt *offsetArray'.
	self var: 'ptr' declareC: 'char **ptr'.
	ptr _ self cCoerce: flattenedArrayOfStrings to: 'char **'.
	idx _ 0.
	[idx < count]
		whileTrue:
			[ptr
				at: idx
				put: (flattenedArrayOfStrings + (interpreterProxy integerValueOf: (offsetArray at: idx))).
			idx _ idx + 1]! !

!OSProcessPlugin methodsFor: 'private' stamp: 'dtl 7/12/2005 22:57'!
primitiveFixPointersInArrayOfStrings
	"This primitive exists only for purposes of testing the
	fixPointersInArrayOfStrings:withOffsets:count: method. I believe it to be
	reasonably machine and compiler independent, but have no way of verifying
	this on a variety of machines, so I'll leave this test method here in case
	someone runs into problems on other hardware or compilers. -dtl"

	| count offsetArray offsets cStringArray flattenedArrayOfStrings |
	self export: true.
	self var: 'offsets' declareC: 'sqInt *offsets'.
	self var: 'flattenedArrayOfStrings' declareC: 'char *flattenedArrayOfStrings'.
	count _ interpreterProxy stackIntegerValue: 0.
	offsetArray _ interpreterProxy stackObjectValue: 1.
	cStringArray _ interpreterProxy stackObjectValue: 2.
	offsets _ interpreterProxy firstIndexableField: offsetArray.
	flattenedArrayOfStrings _ interpreterProxy arrayValueOf: cStringArray.
	self
		fixPointersInArrayOfStrings: flattenedArrayOfStrings
		withOffsets: offsets
		count: count.
	interpreterProxy pop: 4; push: cStringArray! !

!OSProcessPlugin methodsFor: 'private' stamp: 'dtl 2/21/2002 16:33'!
stringFromCString: aCString
	"Answer a new String copied from a null-terminated C string.
	Caution: This may invoke the garbage collector."

	| len newString stringPtr |
	self var: 'aCString' declareC: 'char *aCString'.
	self var: 'stringPtr' declareC: 'char *stringPtr'.
	len _ self cCode: 'strlen(aCString)' inSmalltalk: [0].
	newString _ interpreterProxy
		instantiateClass: interpreterProxy classString
		indexableSize: len.
	stringPtr _ interpreterProxy arrayValueOf: newString.
	self cCode: '(char *)strncpy(stringPtr, aCString, len)'.
	^ newString

! !

!OSProcessPlugin methodsFor: 'private' stamp: 'dtl 8/3/2003 14:55'!
transientCStringFromString: aString
	"Answer a new null-terminated C string copied from aString.
	The string is allocated in object memory, and will be moved
	without warning by the garbage collector. Any C pointer
	reference the the result is valid only until the garbage
	collector next runs. Therefore, this method should only be used
	within a single primitive in a section of code in which the
	garbage collector is guaranteed not to run. Note also that
	this method may itself invoke the garbage collector prior
	to allocating the new C string.

	Warning: The result of this method will be invalidated by the
	next garbage collection, including a GC triggered by creation
	of a new object within a primitive. Do not call this method
	twice to obtain two string pointers."

	| len stringPtr newString cString |
	self returnTypeC: 'char *'.
	self var: 'stringPtr' declareC: 'char *stringPtr'.
	self var: 'cString' declareC: 'char *cString'.
	len _ interpreterProxy sizeOfSTArrayFromCPrimitive: (interpreterProxy arrayValueOf: aString).
	"Allocate space for a null terminated C string."
	interpreterProxy pushRemappableOop: aString.
	newString _ interpreterProxy
		instantiateClass: interpreterProxy classString
		indexableSize: len + 1.
	stringPtr _ interpreterProxy arrayValueOf: interpreterProxy popRemappableOop.
	cString _ interpreterProxy arrayValueOf: newString.		"Point to the actual C string."
	self cCode: '(char *)strncpy(cString, stringPtr, len)'.		"Make a copy of the string."
	cString at: (len) put: 0.									"Null terminate the C string."
	^ cString
! !

!OSProcessPlugin methodsFor: 'private - security' stamp: 'dtl 3/3/2002 08:44'!
sandboxSecurity
	"Answer 1 if running in secure mode, else 0. The osprocessSandboxSecurity
	variable is initialized to -1. On the first call to this method, set its value to
	either 0 (user has full access to the plugin) or 1 (user is not permitted to do
	dangerous things)."

	(osprocessSandboxSecurity < 0) ifTrue: [osprocessSandboxSecurity _ self securityHeurisitic].
	^ osprocessSandboxSecurity
! !

!OSProcessPlugin methodsFor: 'private - security' stamp: 'dtl 9/18/2005 12:37'!
securityHeurisitic
	"Answer 0 to permit full access to OSProcess functions, or 1 if access should be
	restricted for dangerous functions. The rules are:
		- If the security plugin is not present, grant full access
		- If the security plugin can be loaded, restrict access unless user has all
		  of secCanWriteImage, secHasFileAccess and secHasSocketAccess"

	"FIXME: This function has not been tested. -dtl"

	| sCWIfn sHFAfn sHSAfn canWriteImage hasFileAccess hasSocketAccess |
	"If the security plugin can be loaded, use it to check. If not, assume it's ok"

	self var: 'sCWIfn' declareC: 'void (*sCWIfn)(void)'.
	self var: 'sHFAfn' declareC: 'void (*sHFAfn)(void)'.
	self var: 'sHSAfn' declareC: 'void (*sHSAfn)(void)'.

	sCWIfn _ interpreterProxy ioLoadFunction: 'secCanWriteImage' From: 'SecurityPlugin'.
	(sCWIfn = 0) ifTrue: [^ 0 "Cannot find plugin, so grant default full access"].
	canWriteImage _ self cCode: ' ((int (*) (void)) sCWIfn)()'
				inSmalltalk: [1].
	sHFAfn _ interpreterProxy ioLoadFunction: 'secHasFileAccess' From: 'SecurityPlugin'.
	(sHFAfn = 0) ifTrue: [^ 0 "Cannot find plugin, so grant default full access"].
	hasFileAccess _ self cCode: ' ((int (*) (void)) sHFAfn)()'
				inSmalltalk: [1].
	sHSAfn _ interpreterProxy ioLoadFunction: 'secHasSocketAccess' From: 'SecurityPlugin'.
	(sHSAfn = 0) ifTrue: [^ 0 "Cannot find plugin, so grant default full access"].
	hasSocketAccess _ self cCode: ' ((int (*) (void)) sHSAfn)()'
				inSmalltalk: [1].
	((canWriteImage and: [hasFileAccess]) and: [hasSocketAccess])
		ifTrue: [^ 0 "Grant full access"]
		ifFalse: [^ 1 "Restricted access"]
! !

!OSProcessPlugin methodsFor: 'private - conversion' stamp: 'dtl 7/14/2003 19:45'!
copyBytesFrom: charArray1 to: charArray2 length: len

	"| chars |
	chars _ 'abcd'.
	self new copyBytesFrom: 'wxyz' to: chars length: 4.
	chars"

	| p1 p2 idx |
	self var: 'charArray1' type: 'unsigned char *'.
	self var: 'charArray2' type: 'unsigned char *'.
	self var: 'p1' declareC: 'unsigned char *p1'.
	self var: 'p2' declareC: 'unsigned char *p2'.
	self cCode: 'p1 = charArray1' inSmalltalk: [p1 _ 1].
	self cCode: 'p2 = charArray2' inSmalltalk: [p2 _ 1].
	idx _ 0.
	[idx < len] whileTrue:
		[self cCode: '*p2 = *p1'
			inSmalltalk: [charArray2 at: p2 put: (charArray1 at: p1)].
		p1 _ p1 + 1.
		p2 _ p2 + 1.
		idx _ idx + 1]
! !

!OSProcessPlugin methodsFor: 'private - conversion' stamp: 'dtl 3/2/2002 18:44'!
pointerFrom: aByteArray
	"Answer the pointer represented by aByteArray."

	| ptr pointerUnion idx |
	self returnTypeC: 'void *'.
	self var: 'ptr' declareC: 'unsigned char *ptr'.
	self var: 'pointerUnion'
		declareC: 'union {void *address; unsigned char bytes[sizeof(void *)];} pointerUnion'.
	((interpreterProxy isBytes: aByteArray) and:
		[(interpreterProxy stSizeOf: aByteArray) = self sizeOfPointer])
		ifFalse: [^ nil].
	ptr _ interpreterProxy arrayValueOf: aByteArray.
	idx _ 0.
	[idx < self sizeOfPointer]	whileTrue:
		[self cCode: 'pointerUnion.bytes[idx] = ptr[idx]'.
		idx _ idx + 1].
	^ self cCode: 'pointerUnion.address' inSmalltalk: [pointerUnion]
! !

!OSProcessPlugin methodsFor: 'private - conversion' stamp: 'dtl 3/3/2002 10:55'!
sessionIdentifierFrom: aByteArray
	"Answer a session ID represented by aByteArray. The session ID is used in
	the SQFile structure. If that data structure changes, we should see compiler
	warnings about type mismatch with SESSIONIDENTIFIERTYPE."

	| session sessionUnion idx |
	self returnTypeC: 'SESSIONIDENTIFIERTYPE'.
	self var: 'session' declareC: 'unsigned char *session'.
	self var: 'sessionUnion'
		declareC: 'union {SESSIONIDENTIFIERTYPE session; unsigned char bytes[sizeof(SESSIONIDENTIFIERTYPE)];} sessionUnion'.
	((interpreterProxy isBytes: aByteArray) and:
		[(interpreterProxy stSizeOf: aByteArray) = self sizeOfSession])
		ifFalse: [^ nil].
	session _ interpreterProxy arrayValueOf: aByteArray.
	idx _ 0.
	[idx < self sizeOfSession]	whileTrue:
		[self cCode: 'sessionUnion.bytes[idx] = session[idx]'.
		idx _ idx + 1].
	^ self cCode: 'sessionUnion.session' inSmalltalk: [sessionUnion]
! !

!OSProcessPlugin methodsFor: 'private - conversion' stamp: 'dtl 7/14/2003 06:31'!
sizeOfInt
	"Size in bytes of an integer, for this C compiler on this machine."

	^ self cCode: 'sizeof(int)' inSmalltalk: [4]
! !

!OSProcessPlugin methodsFor: 'private - conversion' stamp: 'dtl 7/14/2003 06:32'!
sizeOfPointer
	"Size in bytes of a void pointer, for this C compiler on this machine."

	^ self cCode: 'sizeof(void *)' inSmalltalk: [4].
! !

!OSProcessPlugin methodsFor: 'private - conversion' stamp: 'dtl 3/3/2002 10:53'!
sizeOfSession
	"Size of a SESSIONIDENTIFIERTYPE. Should match usage in the SQFile data structure,
	otherwise we should get compiler warnings."

	^ self cCode: 'sizeof(SESSIONIDENTIFIERTYPE)' inSmalltalk: [4]
! !

!OSProcessPlugin methodsFor: 'private - SQFile access' stamp: 'dtl 2/21/2002 16:35'!
fileRecordSize
	"Answer the size of a SQFile data structure in bytes."

	^ self cCode: 'sizeof(SQFile)' inSmalltalk: [20]
! !

!OSProcessPlugin methodsFor: 'private - SQFile access' stamp: 'dtl 3/1/2002 14:10'!
fileValueOf: anSQFileRecord
	"Return a pointer to the first byte of of the SQFile data structure file record within
	anSQFileRecord, which is expected to be a ByteArray of size self>>fileRecordSize."

	self returnTypeC: 'SQFile *'.
	((interpreterProxy isBytes: anSQFileRecord) and: [
		(interpreterProxy byteSizeOf: anSQFileRecord) == self fileRecordSize])
			ifTrue:
				[^ interpreterProxy arrayValueOf: anSQFileRecord]
			ifFalse:
				[interpreterProxy primitiveFail.
				^ nil]

! !

!OSProcessPlugin methodsFor: 'private - SQFile access' stamp: 'dtl 9/18/2002 10:44'!
isSQFileObject: objectPointer
	"Answer true if objectPointer appears to be a valid SQFile ByteArray. This check
	is appropriate if objectPointer has been passed as a parameter to a primitive, and
	is expected to represent a valid file reference."

	| idx sqFileBytes |
	self var: 'sqFileBytes' declareC: 'unsigned char *sqFileBytes'.
	(((interpreterProxy isBytes: objectPointer) and:
		 [(interpreterProxy byteSizeOf: objectPointer) = self fileRecordSize]))
			ifFalse: [^ false].
	"Check for the common failure mode of a SQFile record with all zeros."
	sqFileBytes _ interpreterProxy arrayValueOf: objectPointer.
	idx _ 0.
	[idx < self fileRecordSize]
		whileTrue:
			[((sqFileBytes at: idx) ~= 0) ifTrue: [^ true].
			idx _ idx + 1].
	^ false
! !

!OSProcessPlugin methodsFor: 'private - SQFile access' stamp: 'dtl 2/21/2002 16:35'!
newSQFileByteArray
	"Answer a new ByteArray sized to contain a SQFile data structure."

	^ interpreterProxy
		instantiateClass: interpreterProxy classByteArray
		indexableSize: self fileRecordSize
! !

!OSProcessPlugin methodsFor: 'private - SQSocket access' stamp: 'dtl 2/21/2002 16:36'!
newSQSocketByteArray
	"Answer a new ByteArray sized to contain a SQSocket data structure."

	^ interpreterProxy
		instantiateClass: interpreterProxy classByteArray
		indexableSize: self socketRecordSize
! !

!OSProcessPlugin methodsFor: 'private - SQSocket access' stamp: 'dtl 2/21/2002 16:36'!
socketRecordSize
	"Answer the size of a SQSocket data structure in bytes."

	^ self cCode: 'sizeof(SQSocket)' inSmalltalk: [12]
! !

!OSProcessPlugin methodsFor: 'private - SQSocket access' stamp: 'dtl 2/22/2002 15:58'!
socketValueOf: anSQSocketRecord
	"Return a pointer to the first byte of of the SQsocket data structure socket record within
	anSQSocketRecord, which is expected to be a ByteArray of size self>>socketRecordSize."

	self returnTypeC: 'SQSocket *'.
	((interpreterProxy isBytes: anSQSocketRecord) and: [
		(interpreterProxy byteSizeOf: anSQSocketRecord) == self socketRecordSize])
			ifTrue: [^ interpreterProxy arrayValueOf: anSQSocketRecord]
			ifFalse: [^ nil]

! !

!OSProcessPlugin methodsFor: 'deprecated' stamp: 'dtl 6/13/2005 00:43'!
primitiveMakePipe
	"Create a pipe, and answer an array of two file handles for the pipe writer and reader.
	The readerIOStream and writerIOStream variables represent the low level pipe streams,
	which will be of type (FILE *) or HANDLE, depending on what the FilePlugin support
	code is using to represent file streams. FILEHANDLETYPE is defined in my subclasses
	in the #declareCVarsIn: class method."

	| reader writer readerPtr writerPtr readerIOStream writerIOStream
	readerIOStreamPtr writerIOStreamPtr thisSession arrayResult |
	self export: true.
	self var: 'readerIOStream' declareC: 'FILEHANDLETYPE readerIOStream'.
	self var: 'writerIOStream' declareC: 'FILEHANDLETYPE writerIOStream'.
	self var: 'readerIOStreamPtr' declareC: 'FILEHANDLETYPE *readerIOStreamPtr'.
	self var: 'writerIOStreamPtr' declareC: 'FILEHANDLETYPE *writerIOStreamPtr'.
	self var: 'readerPtr' declareC: 'SQFile *readerPtr'.
	self var: 'writerPtr' declareC: 'SQFile *writerPtr'.
	self var: 'arrayPtr' declareC: 'int *arrayPtr'.
	self var: 'thisSession' type: 'SESSIONIDENTIFIERTYPE'.
	thisSession _ self getThisSessionIdentifier.

	"Create the anonymous OS pipe"
	self cCode: 'readerIOStreamPtr = &readerIOStream'.
	self cCode: 'writerIOStreamPtr = &writerIOStream'.
	(self makePipeForReader: readerIOStreamPtr writer: writerIOStreamPtr)
		ifFalse: [^ interpreterProxy primitiveFail].

	"Map the writer end to a SQFile data structure"
	writer _ self newSQFileByteArray.
	writerPtr _ self fileValueOf: writer.
	self cCode: 'writerPtr->file = writerIOStream'.
	self cCode: 'writerPtr->sessionID = thisSession'.
	self cCode: 'writerPtr->writable = 1'.
	self cCode: 'writerPtr->lastOp = 0'.
	interpreterProxy pushRemappableOop: writer.

	"Map the reader end to a SQFile data structure"
	reader _ self newSQFileByteArray.
	readerPtr _ self fileValueOf: reader.
	self cCode: 'readerPtr->file = readerIOStream'.	
	self cCode: 'readerPtr->sessionID = thisSession'.
	self cCode: 'readerPtr->writable = 0'.
	self cCode: 'readerPtr->lastOp = 0'.
	interpreterProxy pushRemappableOop: reader.

	"Answer an array of two SQFile pipe ends"
	arrayResult _ interpreterProxy instantiateClass: interpreterProxy classArray indexableSize: 2.
	interpreterProxy stObject: arrayResult at: 1 put: interpreterProxy popRemappableOop.	"reader"
	interpreterProxy stObject: arrayResult at: 2 put: interpreterProxy popRemappableOop. "writer"
	interpreterProxy pop: 1; push: arrayResult
! !

!OSProcessPlugin methodsFor: 'deprecated' stamp: 'dtl 6/13/2005 00:43'!
primitiveMakePipeWithSessionIdentifier
	"Create a pipe, and answer an array of two file handles for the pipe writer and reader.
	The session identifier is passed as the parameter to this primitive. Use this variant
	if the session identifier is not available directly in the VM (as may be the case if
	it is not possible to link from this plugin to a variable elsewhere in the VM).
	The readerIOStream and writerIOStream variables represent the low level pipe streams,
	which will be of type (FILE *) or HANDLE, depending on what the FilePlugin support
	code is using to represent file streams. FILEHANDLETYPE is defined in my subclasses
	in the #declareCVarsIn: class method."

	| reader writer readerPtr writerPtr readerIOStream writerIOStream
	readerIOStreamPtr writerIOStreamPtr thisSession arrayResult |
	self export: true.
	self var: 'readerIOStream' declareC: 'FILEHANDLETYPE readerIOStream'.
	self var: 'writerIOStream' declareC: 'FILEHANDLETYPE writerIOStream'.
	self var: 'readerIOStreamPtr' declareC: 'FILEHANDLETYPE *readerIOStreamPtr'.
	self var: 'writerIOStreamPtr' declareC: 'FILEHANDLETYPE *writerIOStreamPtr'.
	self var: 'readerPtr' declareC: 'SQFile *readerPtr'.
	self var: 'writerPtr' declareC: 'SQFile *writerPtr'.
	self var: 'arrayPtr' declareC: 'int *arrayPtr'.
	self var: 'thisSession' type: 'SESSIONIDENTIFIERTYPE'.
	thisSession _ self sessionIdentifierFrom: (interpreterProxy stackObjectValue: 0).

	"Create the anonymous OS pipe"
	self cCode: 'readerIOStreamPtr = &readerIOStream'.
	self cCode: 'writerIOStreamPtr = &writerIOStream'.
	(self makePipeForReader: readerIOStreamPtr writer: writerIOStreamPtr)
		ifFalse: [^ interpreterProxy primitiveFail].

	"Map the writer end to a SQFile data structure"
	writer _ self newSQFileByteArray.
	writerPtr _ self fileValueOf: writer.
	self cCode: 'writerPtr->file = writerIOStream'.
	self cCode: 'writerPtr->sessionID = thisSession'.
	self cCode: 'writerPtr->writable = 1'.
	self cCode: 'writerPtr->lastOp = 0'.
	interpreterProxy pushRemappableOop: writer.

	"Map the reader end to a SQFile data structure"
	reader _ self newSQFileByteArray.
	readerPtr _ self fileValueOf: reader.
	self cCode: 'readerPtr->file = readerIOStream'.	
	self cCode: 'readerPtr->sessionID = thisSession'.
	self cCode: 'readerPtr->writable = 0'.
	self cCode: 'readerPtr->lastOp = 0'.
	interpreterProxy pushRemappableOop: reader.

	"Answer an array of two SQFile pipe ends"
	arrayResult _ interpreterProxy instantiateClass: interpreterProxy classArray indexableSize: 2.
	interpreterProxy stObject: arrayResult at: 1 put: interpreterProxy popRemappableOop.	"reader"
	interpreterProxy stObject: arrayResult at: 2 put: interpreterProxy popRemappableOop. "writer"
	interpreterProxy pop: 2; push: arrayResult
! !


!OSProcessPlugin class methodsFor: 'translation' stamp: 'dtl 3/5/2005 11:55'!
concreteClass

	"OSProcessPlugin concreteClass"

	| numericOsVersion |
	(self platformName = 'Win32') ifTrue: [^ Win32OSProcessPlugin concreteClass].
	(self platformName = 'unix') ifTrue: [^ UnixOSProcessPlugin concreteClass].
	numericOsVersion _ self osVersion asInteger ifNil: [0].
	((self platformName = 'Mac OS') and: [numericOsVersion >= 1000])
		ifTrue: [^ UnixOSProcessPlugin concreteClass].
	self inform: 'no concrete class available for ', self platformName.
	^ nil
! !

!OSProcessPlugin class methodsFor: 'translation' stamp: 'dtl 9/15/2005 06:28'!
declareCVarsIn: cg

	"Additional declarations are done in my subclasses. This permits a meaningful
	module identification string to be written with the name of the actual subclass
	being generated."

	cg addHeaderFile: '"FilePlugin.h"'.
	cg addHeaderFile: '"SocketPlugin.h"
#define SESSIONIDENTIFIERTYPE int'.	"Data type for sessionID, should generate compiler warnings
									if someone changes the data type elsewhere in the VM."
	cg var: 'osprocessSandboxSecurity' declareC: 'int osprocessSandboxSecurity'.
! !

!OSProcessPlugin class methodsFor: 'translation' stamp: 'dtl 8/8/2005 06:10'!
idString

	^ 'D T Lewis 2005'! !

!OSProcessPlugin class methodsFor: 'translation' stamp: 'dtl 2/21/2002 15:57'!
shouldBeTranslated
	"Translate one of my subclasses instead."

	^ false! !

!OSProcessPlugin class methodsFor: 'documentation' stamp: 'dtl 3/3/2002 13:14'!
about

	"OSProcessPlugin about"

	^ StringHolder new contents: self howToInstall; openLabel: 'about ', self asString! !

!OSProcessPlugin class methodsFor: 'documentation' stamp: 'dtl 3/5/2005 12:06'!
howToInstall

	"OSProcessPlugin about"

	| installedPluginString |
	installedPluginString _ self concreteClass installedModuleVersionString.
	^ (installedPluginString isNil
		ifTrue:
			['Your system does not have an installed plugin module. Your current configuration is ']
		ifFalse:
			[(self versionString = installedPluginString)
		ifTrue: ['Your system currently has an installed, up to date plugin module. Your current configuration is ']
		ifFalse: ['Your system currently has an installed plugin module, but its version does not match the OSProcessPlugin change set in your image, so you may wish to rebuild it. Your current configuration is ']]), self installedModule, '.

', self name, ' is a plugin written in Smalltalk. The complete source code for ', self name, ' for your platform is contained in class OSProcessPlugin and its subclasses. No external support code is required. It is intended to be translated to C and compiled as a shared object module (or linked directly with the VM if built as an internal plugin). ', self name, ' has a number of subclasses which allow platform specific methods to be generated for Unix, Linux and Windows platforms. The concrete subclass for your ', self platformName, ' Squeak is ', self concreteClass printString, '.

If OSProcess cannot connect to the operating system ("OSProcess accessor canAccessSystem" evaluates to false), then you need to build and install the ', self name,' plugin.

Use VMMaker to generate the C source code for ', self name, '. The correct ', self name, ' subclass will automatically appear in the list of available modules. VMMaker will create a C source code file called ', self concreteClass moduleName, '.c in the generated source code tree. Compile the plugin according to the instructions provided with your platform source distribution.

If ', self name, ' is built as an external plugin, the C source file will be compiled and linked into a shared module called UnixOSProcessPlugin.so (for Unix or Linux) or Win32OSProcessPlugin.dll (for Windows). If the plugin is built internally, it will be linked directly into the Squeak VM executable. The shared module file should be moved to a place where Squeak can find it (your Squeak home directory is fine). After restarting Squeak, OSProcess will begin functioning with access to the operating system.'! !

!OSProcessPlugin class methodsFor: 'concrete subclasses' stamp: 'dtl 8/7/2005 22:06'!
hasInterpreterGetThisSession
	"True if the Interpreter has been updated to support #getThisSession.
	If false, we need to rely on linking to a global variable, which in turn
	requires that the FilePlugin module be built internally."

	^ Interpreter canUnderstand: #getThisSessionID! !

!OSProcessPlugin class methodsFor: 'concrete subclasses' stamp: 'dtl 2/23/2002 06:14'!
hasThisSessionAccessorInFilePlugin
	"True if the FilePlugin has been updated to support FilePlugin>>getThisSession.
	If false, we need to rely on linking to a global variable, which in turn
	requires that the FilePlugin module be built internally."

	^ (Smalltalk hasClassNamed: 'FilePlugin') and: [FilePlugin canUnderstand: #getThisSession]! !

!OSProcessPlugin class methodsFor: 'version testing' stamp: 'dtl 8/13/2003 21:58'!
canAccessSystem
	"True if an installed OSPP is available and can access the system"

	"OSProcessPlugin concreteClass canAccessSystem"

	^ self concreteClass primGetPid notNil
! !

!OSProcessPlugin class methodsFor: 'version testing' stamp: 'dtl 3/1/2002 17:26'!
installedModuleName
	"Answer the module name string for the plugin which is actually running in the VM"

	"OSProcessPlugin installedModuleName"

	^ self concreteClass installedModuleName! !

!OSProcessPlugin class methodsFor: 'version testing' stamp: 'dtl 3/1/2002 17:25'!
installedModuleVersionString
	"Answer the version string for the plugin which is actually running in the VM"

	"OSProcessPlugin installedModuleVersionString"

	^ self concreteClass installedModuleVersionString! !

!OSProcessPlugin class methodsFor: 'version testing' stamp: 'dtl 8/13/2003 21:56'!
primGetPid
	"Answer the OS process ID for the OS process in which I am currently executing."

	^ self subclassResponsibility! !

!OSProcessPlugin class methodsFor: 'version identification' stamp: 'dtl 3/5/2005 11:54'!
versionInformation

	"OSProcessPlugin versionInformation"

	| osp ospString |
	osp _ Smalltalk at: #OSProcess ifAbsent: [].
	ospString _ osp isNil
		ifTrue:
			[' (OSProcess not installed in this image)']
		ifFalse:
			[' installed with OSProcess version ', osp versionString].
	^ self vmVersion, ' with ', self concreteClass moduleName, ' version ',
		self versionString, ' using concrete class ', self concreteClass name,
		ospString, ', running on ', ((self getSystemAttribute: 1002) asString),
		' ', self platformName, ' VM with installed plugin ',
		(self canAccessSystem
			ifTrue:
				[self concreteClass installedModuleName isNil
					ifTrue:
						['(old version of plugin cannot identify itself)']
					ifFalse:
						[self concreteClass installedModuleName, ' version ',
						self installedModuleVersionString,
						((self versionString = self concreteClass installedModuleVersionString)
							ifTrue: [' (plugin matches class in image)']
							ifFalse: [' (plugin does not match class in image)'])]]
			ifFalse:
				['(OSProcessPlugin installed module not accessible)'])

! !

!OSProcessPlugin class methodsFor: 'version identification' stamp: 'dtl 11/5/2005 07:51'!
versionString
	"Version numbering is independent of OSProcess class>>versionString, and should be
	changed only when the plugin changes. Version numbers will begin arbitrarily
	with 2.1 as of OSProcess 2.7. See the instance side version of this method for the
	implementation.

	Important: When this method is changed, the instance side method must also be
	changed to match."

	"OSProcessPlugin versionString"

	"4.0 supports 64bit code base"

	^ '4.0.1'
! !

!OSProcessPlugin class methodsFor: 'version dependent' stamp: 'dtl 3/5/2005 11:51'!
getSystemAttribute: attributeID 
	"After Squeak version 3.6, #getSystemAttribute was moved to SmalltalkImage "

	^ ((Smalltalk classNamed: 'SmalltalkImage')
		ifNil: [^ Smalltalk getSystemAttribute: attributeID]) current getSystemAttribute: attributeID! !

!OSProcessPlugin class methodsFor: 'version dependent' stamp: 'dtl 3/5/2005 11:55'!
osVersion
	"After Squeak version 3.6, #osVersion was moved to SmalltalkImage "

	^ ((Smalltalk classNamed: 'SmalltalkImage')
		ifNil: [^ Smalltalk osVersion]) current osVersion! !

!OSProcessPlugin class methodsFor: 'version dependent' stamp: 'dtl 3/5/2005 11:53'!
platformName 
	"After Squeak version 3.6, #platformName was moved to SmalltalkImage "

	^ ((Smalltalk classNamed: 'SmalltalkImage')
		ifNil: [^ Smalltalk platformName]) current platformName! !

!OSProcessPlugin class methodsFor: 'version dependent' stamp: 'dtl 3/5/2005 11:52'!
vmVersion
	"After Squeak version 3.6, #vmVersion was moved to SmalltalkImage "

	^ ((Smalltalk classNamed: 'SmalltalkImage')
		ifNil: [^ Smalltalk vmVersion]) current vmVersion! !

!OSProcessPlugin class methodsFor: 'private - testing' stamp: 'dtl 6/12/2005 23:40'!
primFixPointersInArrayOfStrings: cStringArray withOffsets: offsetArray count: count
	"This primitive call exists only for purposes of testing the
	fixPointersInArrayOfStrings:withOffsets:count: method. I believe it to be
	reasonably machine and compiler independent, but have no way of verifying
	this on a variety of machines, so I'll leave this test method here in case
	someone runs into problems on other hardware or compilers. -dtl"

	"| a |
	a _ OSProcess thisOSProcess envAsFlatArrayAndOffsets: UnixProcess env.
	OSProcessPlugin primFixPointersInArrayOfStrings: (a at: 1)
		withOffsets: (a at: 2)
		count: (a at: 2) size"

	<primitive: 'primitiveFixPointersInArrayOfStrings' module: 'UnixOSProcessPlugin'>
	^ nil! !

!OSProcessPlugin class methodsFor: 'deprecated' stamp: 'dtl 8/8/2003 23:03'!
installedModule
	"Deprecated - this method name is misleading. Replaced as of OSP 3.1"

	^ self versionInformation! !


!UnixOSProcessPlugin methodsFor: 'initialize-release' stamp: 'dtl 8/3/2002 23:41'!
shutdownModule

	self export: true.
	self restoreDefaultSignalHandlers! !

!UnixOSProcessPlugin methodsFor: 'initialize-release' stamp: 'dtl 1/31/2004 19:45'!
signalArraySize
	"Number of possible signals for this OS plus one. The signal handler arrays
	declared in #declareCVarsIn: are this size."

	^ self cCode: 'NSIG' inSmalltalk: [65]
! !

!UnixOSProcessPlugin methodsFor: 'accessing' stamp: 'dtl 8/3/2002 19:32'!
originalSignalHandlers
	"An array of signal handler function addresses, one for each signal type. The
	value at each entry is the address of the original signal handler function prior
	to having set a handler."

	self returnTypeC: 'void **'.
	^ originalSigHandlers! !

!UnixOSProcessPlugin methodsFor: 'accessing' stamp: 'dtl 8/3/2002 19:36'!
semaphoreIndices
	"An array of Smalltalk Semaphore indices, one for each signal type for which
	a handler has been set. If an entry is zero, then no handler is active."

	self returnTypeC: 'unsigned char *'.
	^ semaIndices! !

!UnixOSProcessPlugin methodsFor: 'accessing' stamp: 'dtl 8/3/2002 19:31'!
signalHandlers
	"An array of signal handler function addresses, with each entry corresponding
	to a signal type."

	self returnTypeC: 'void **'.
	^ sigHandlers! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl 2/21/2002 15:40'!
primitiveFileProtectionMask
	"Call stat(2) to obtain the file protection mask for a file. Answer an Array of
	four integers representing the protection mask, or answer errno on failure. The
	protection mask is four Integers, each of which may be considered an octal digit
	(0-7), with bit values 4, 2, and 1. The first digit selects the set user ID (4) and set
	group ID (2) and save text image (1) attributes. The second digit selects permissions
	for the user who owns the file: read (4), write (2), and execute (1); the third
	selects permissions for other users in the file's group, with the same values; and
	the fourth for other users not in the file's group, with the same values."

	| path statBuf buffer mode errno result |
	self export: true.
	self var: 'path' declareC: 'char *path'.
	self var: 'statBuf' declareC: 'struct stat *statBuf'.
	self var: 'errno' declareC: 'extern int errno'.

	buffer _ interpreterProxy
				instantiateClass: interpreterProxy classByteArray
				indexableSize: (self cCode: 'sizeof(struct stat)' inSmalltalk: [1000]).
	statBuf _ interpreterProxy arrayValueOf: buffer.
	path _ self transientCStringFromString: (interpreterProxy stackObjectValue: 0).
	mode _ (self cCode: 'stat(path, statBuf)' inSmalltalk: [0]).
	(mode == 0) ifTrue:
			[mode _ self cCode: 'statBuf->st_mode' inSmalltalk: [0].
			result _ interpreterProxy instantiateClass: interpreterProxy classArray indexableSize: 4.
			interpreterProxy
				stObject: result
				at: 4
				put: (interpreterProxy integerObjectOf: (self cCode: 'mode & 07' inSmalltalk: [0])).
			interpreterProxy
				stObject: result
				at: 3
				put: (interpreterProxy integerObjectOf: (self cCode: '(mode & 070) >> 3' inSmalltalk: [0])).
			interpreterProxy
				stObject: result
				at: 2
				put: (interpreterProxy integerObjectOf: (self cCode: '(mode & 0700) >> 6' inSmalltalk: [0])).
			interpreterProxy
				stObject: result
				at: 1
				put: (interpreterProxy integerObjectOf: (self cCode: '(mode & 07000) >> 9' inSmalltalk: [0])).
			interpreterProxy pop: 2; push: result]
		ifFalse:
			[interpreterProxy pop: 2; pushInteger: errno].

! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl 2/21/2002 15:40'!
primitiveFileStat
	"Call stat(2) to obtain the file protection mask for a file. Answer errno on failure,
	or on success answer an array with: UID with: GID with: protectionMask. The	
	protectionMask is an Array of four integers representing the protection mask, or
	answer errno on failure. The protection mask is four Integers, each of which may
	be considered an octal digit (0-7), with bit values 4, 2, and 1. The first digit selects
	the set user ID (4) and set group ID (2) and save text image (1) attributes. The second
	digit selects permissions for the user who owns the file: read (4), write (2), and
	execute (1); the third selects permissions for other users in the file's group, with
	the same values; and the fourth for other users not in the file's group, with the
	same values."

	| path statBuf buffer uid gid mask mode errno result |
	self export: true.
	self var: 'path' declareC: 'char *path'.
	self var: 'statBuf' declareC: 'struct stat *statBuf'.
	self var: 'errno' declareC: 'extern int errno'.

	result _ interpreterProxy
				instantiateClass: interpreterProxy classArray
				indexableSize: 3.
	uid _ interpreterProxy
				instantiateClass: interpreterProxy classByteArray
				indexableSize: (self cCode: 'sizeof(uid_t)' inSmalltalk: [4]).
	gid _ interpreterProxy
				instantiateClass: interpreterProxy classByteArray
				indexableSize: (self cCode: 'sizeof(gid_t)' inSmalltalk: [4]).
	mask _ interpreterProxy
				instantiateClass: interpreterProxy
				classArray indexableSize: 4.
	buffer _ interpreterProxy
				instantiateClass: interpreterProxy classByteArray
				indexableSize: (self cCode: 'sizeof(struct stat)' inSmalltalk: [1000]).
	statBuf _ interpreterProxy arrayValueOf: buffer.
	path _ self transientCStringFromString: (interpreterProxy stackObjectValue: 0).
	mode _ (self cCode: 'stat(path, statBuf)' inSmalltalk: [0]).
	(mode == 0) ifTrue:
			[mode _ self cCode: 'statBuf->st_mode' inSmalltalk: [0].
			interpreterProxy
				stObject: mask
				at: 4
				put: (interpreterProxy integerObjectOf: (self cCode: 'mode & 07' inSmalltalk: [0])).
			interpreterProxy
				stObject: mask
				at: 3
				put: (interpreterProxy integerObjectOf: (self cCode: '(mode & 070) >> 3' inSmalltalk: [0])).
			interpreterProxy
				stObject: mask
				at: 2
				put: (interpreterProxy integerObjectOf: (self cCode: '(mode & 0700) >> 6' inSmalltalk: [0])).
			interpreterProxy
				stObject: mask
				at: 1
				put: (interpreterProxy integerObjectOf: (self cCode: '(mode & 07000) >> 9' inSmalltalk: [0])).
			interpreterProxy
				stObject: result
				at: 3
				put: mask.
			interpreterProxy
				stObject: result
				at: 2
				put: (interpreterProxy integerObjectOf: (self cCode: 'statBuf->st_gid' inSmalltalk: [0])).
			interpreterProxy
				stObject: result
				at: 1
				put: (interpreterProxy integerObjectOf: (self cCode: 'statBuf->st_uid' inSmalltalk: [0])).
			interpreterProxy pop: 2; push: result]
		ifFalse:
			[interpreterProxy pop: 2; pushInteger: errno]
! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl 6/13/2005 00:23'!
primitiveGetStdErrHandle
	"Answer the file handle for standard error of my OS process"

	| file thisSession fileOop |
	self export: true.
	self var: 'file' declareC: 'SQFile *file'.
	self var: 'thisSession' type: 'SESSIONIDENTIFIERTYPE'.
	fileOop _ self newSQFileByteArray.
	file _ self fileValueOf: fileOop.
	thisSession _ self getThisSessionIdentifier.
	self cCode: 'file->file = stderr'.
	self cCode: 'file->sessionID = thisSession'.
	self cCode: 'file->writable = 1'.
	self cCode: 'file->lastOp = 0'.
	interpreterProxy pop: 1; push: fileOop
! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl 6/13/2005 00:23'!
primitiveGetStdErrHandleWithSessionIdentifier
	"Answer the file handle for standard error of my OS process. The session
	identifier is passed as the parameter to this primitive. Use this variant if
	the session identifier is not available directly in the VM (as may be the
	case if it is not possible to link from this plugin to a variable elsewhere
	in the VM)."

	| file thisSession fileOop |
	self export: true.
	self var: 'file' declareC: 'SQFile *file'.
	self var: 'thisSession' type: 'SESSIONIDENTIFIERTYPE'.
	fileOop _ self newSQFileByteArray.
	file _ self fileValueOf: fileOop.
	thisSession _ self sessionIdentifierFrom: (interpreterProxy stackObjectValue: 0).
	self cCode: 'file->file = stderr'.
	self cCode: 'file->sessionID = thisSession'.
	self cCode: 'file->writable = 1'.
	self cCode: 'file->lastOp = 0'.
	interpreterProxy pop: 2; push: fileOop
! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl 6/13/2005 00:23'!
primitiveGetStdInHandle
	"Answer the file handle for standard input of my OS process"

	| file thisSession fileOop |
	self export: true.
	self var: 'file' declareC: 'SQFile *file'.
	self var: 'thisSession' type: 'SESSIONIDENTIFIERTYPE'.
	fileOop _ self newSQFileByteArray.
	file _ self fileValueOf: fileOop.
	thisSession _ self getThisSessionIdentifier.
	self cCode: 'file->file = stdin'.
	self cCode: 'file->sessionID = thisSession'.
	self cCode: 'file->writable = 0'.
	self cCode: 'file->lastOp = 0'.
	interpreterProxy pop: 1; push: fileOop
! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl 6/13/2005 00:23'!
primitiveGetStdInHandleWithSessionIdentifier
	"Answer the file handle for standard input of my OS process. The session
	identifier is passed as the parameter to this primitive. Use this variant if
	the session identifier is not available directly in the VM (as may be the
	case if it is not possible to link from this plugin to a variable elsewhere
	in the VM)."

	| file thisSession fileOop |
	self export: true.
	self var: 'file' declareC: 'SQFile *file'.
	self var: 'thisSession' type: 'SESSIONIDENTIFIERTYPE'.
	fileOop _ self newSQFileByteArray.
	file _ self fileValueOf: fileOop.
	thisSession _ self sessionIdentifierFrom: (interpreterProxy stackObjectValue: 0).
	self cCode: 'file->file = stdin'.
	self cCode: 'file->sessionID = thisSession'.
	self cCode: 'file->writable = 0'.
	self cCode: 'file->lastOp = 0'.
	interpreterProxy pop: 2; push: fileOop
! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl 6/13/2005 00:23'!
primitiveGetStdOutHandle
	"Answer the file handle for standard output of my OS process"

	| file thisSession fileOop |
	self export: true.
	self var: 'file' declareC: 'SQFile *file'.
	self var: 'thisSession' type: 'SESSIONIDENTIFIERTYPE'.
	fileOop _ self newSQFileByteArray.
	file _ self fileValueOf: fileOop.
	thisSession _ self getThisSessionIdentifier.
	self cCode: 'file->file = stdout'.
	self cCode: 'file->sessionID = thisSession'.
	self cCode: 'file->writable = 1'.
	self cCode: 'file->lastOp = 0'.
	interpreterProxy pop: 1; push: fileOop
! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl 6/13/2005 00:23'!
primitiveGetStdOutHandleWithSessionIdentifier
	"Answer the file handle for standard output of my OS process. The session
	identifier is passed as the parameter to this primitive. Use this variant if
	the session identifier is not available directly in the VM (as may be the
	case if it is not possible to link from this plugin to a variable elsewhere
	in the VM)."

	| file thisSession fileOop |
	self export: true.
	self var: 'file' declareC: 'SQFile *file'.
	self var: 'thisSession' type: 'SESSIONIDENTIFIERTYPE'.
	fileOop _ self newSQFileByteArray.
	file _ self fileValueOf: fileOop.
	thisSession _ self sessionIdentifierFrom: (interpreterProxy stackObjectValue: 0).
	self cCode: 'file->file = stdout'.
	self cCode: 'file->sessionID = thisSession'.
	self cCode: 'file->writable = 1'.
	self cCode: 'file->lastOp = 0'.
	interpreterProxy pop: 2; push: fileOop
! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl 9/18/2002 10:33'!
primitiveIsAtEndOfFile
	"Take a struct SQFile from the stack, and call feof(3) to determine if the file has
	reached end of file."

	| sqFileOop sqFile result |
	self export: true.
	self var: 'sqFile' declareC: 'SQFile *sqFile'.
	sqFileOop _ interpreterProxy stackValue: 0.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	sqFile _ interpreterProxy arrayValueOf: sqFileOop.
	((self cCode: 'feof(sqFile->file)' inSmalltalk: [0]) == 0)
		ifTrue: [result _ interpreterProxy falseObject]
		ifFalse: [result _ interpreterProxy trueObject].
	interpreterProxy pop: 2; push: result

! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl 8/3/2003 15:36'!
primitiveRealpath
	"Answer the real path for a path string as determined by realpath()."

	|  bufferSize pathString newPathString buffer realpathResult s |
	self export: true.
	self var: 'pathString' declareC: 'char *pathString'.
	self var: 'buffer' declareC: 'char *buffer'.
	self var: 'realpathResult' declareC: 'char * realpathResult'.

	bufferSize _ 1024.
	newPathString _ interpreterProxy
		instantiateClass: interpreterProxy classString
		indexableSize: bufferSize.
	interpreterProxy pushRemappableOop: newPathString.
	pathString _ self transientCStringFromString: (interpreterProxy stackObjectValue: 0).
	newPathString _ interpreterProxy popRemappableOop.
	buffer _ interpreterProxy arrayValueOf: newPathString.
	realpathResult _ self cCode: 'realpath(pathString, buffer)' inSmalltalk: [''].
	(realpathResult = 0)
		ifTrue:
			[^ interpreterProxy primitiveFail]
		ifFalse:
			[((self strlen: realpathResult) >= 1024) ifTrue:
				[self perror: 'warning: statically allocated array exceeded in UnixOSProcessPlugin>>primitiveRealPath, object memory may have been corrupted'.
				^ interpreterProxy primitiveFail].
			s _ self stringFromCString: realpathResult.
			interpreterProxy pop: 2; push: s]

! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl 6/13/2005 00:45'!
primitiveSQFileFlush
	"Take a struct SQFile from the stack, and call fflush() to flush the OS stream. This flushes the
	file stream in the C library, not the stream in Smalltalk. For output streams, consider setting
	the OS stream (C library) to unbuffered output, and letting Smalltalk do all the buffering."

	| sqFileOop sqFile thisSession |
	self export: true.
	self var: 'sqFile' declareC: 'SQFile *sqFile'.
	self var: 'thisSession' type: 'SESSIONIDENTIFIERTYPE'.
	sqFileOop _ interpreterProxy stackValue: 0.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	sqFile _ interpreterProxy arrayValueOf: sqFileOop.
	"If the session identifier can be obtained, use it as an additional consistency check"
	thisSession _ self getThisSessionIdentifier.
	thisSession ifNotNil:
		[(thisSession = self cCode: 'sqFile->sessionID' inSmalltalk: [-1])
			ifFalse: [^ interpreterProxy primitiveFail]].
	interpreterProxy pop: 2; pushInteger: (self cCode: 'fflush(sqFile->file)')
! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl 6/13/2005 00:45'!
primitiveSQFileFlushWithSessionIdentifier
	"Take a struct SQFile from the stack, and call fflush() to flush the OS stream. This flushes the
	file stream in the C library, not the stream in Smalltalk. For output streams, consider setting
	the OS stream (C library) to unbuffered output, and letting Smalltalk do all the buffering.
	The session identifier is passed as the parameter to this primitive. Use this variant if the session
	identifier is not available directly in the VM (as may be the case if it is not possible to link from
	this plugin to a variable elsewhere in the VM)."

	| sqFileOop sqFile thisSession |
	self export: true.
	self var: 'sqFile' declareC: 'SQFile *sqFile'.
	self var: 'thisSession' type: 'SESSIONIDENTIFIERTYPE'.
	sqFileOop _ interpreterProxy stackValue: 1.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	sqFile _ interpreterProxy arrayValueOf: sqFileOop.
	thisSession _ self sessionIdentifierFrom: (interpreterProxy stackObjectValue: 0).
	(thisSession = (self cCode: 'sqFile->sessionID' inSmalltalk: [-1]))
		ifFalse: [^ interpreterProxy primitiveFail]
		ifTrue: [interpreterProxy pop: 3; pushInteger: (self cCode: 'fflush(sqFile->file)')]

! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl 6/13/2005 00:45'!
primitiveSQFileSetBlocking
	"Take a struct SQFile from the stack, and call fcntl() to set the file for blocking I/O."

	| sqFile thisSession sqFileOop descriptor flags retVal |
	self export: true.
	self var: 'sqFile' declareC: 'SQFile *sqFile'.
	self var: 'thisSession' type: 'SESSIONIDENTIFIERTYPE'.
	sqFileOop _ interpreterProxy stackValue: 0.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	sqFile _ interpreterProxy arrayValueOf: sqFileOop.
	"If the session identifier can be obtained, use it as an additional consistency check"
	thisSession _ self getThisSessionIdentifier.
	thisSession ifNotNil:
		[(thisSession = self cCode: 'sqFile->sessionID' inSmalltalk: [-1])
			ifFalse: [^ interpreterProxy primitiveFail]].
	descriptor _ self fileDescriptorFrom: sqFileOop.
	(descriptor < 0) ifTrue: [^ interpreterProxy primitiveFail].
	flags _ self cCode: 'fcntl(descriptor, F_GETFL)'.
	retVal _ self cCode: 'fcntl(descriptor, F_SETFL, flags & ~O_NONBLOCK)'.
	interpreterProxy pop: 2; pushInteger: retVal

! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl 6/13/2005 00:46'!
primitiveSQFileSetBlockingWithSessionIdentifier
	"Take a struct SQFile from the stack, and call fcntl() to set the file for blocking I/O.
	Use this variant if the session identifier is not available directly in the VM (as may be
	the case if it is not possible to link from this plugin to a variable elsewhere in the VM)."

	| sqFile sqFileOop descriptor flags retVal thisSession |
	self export: true.
	self var: 'sqFile' declareC: 'SQFile *sqFile'.
	self var: 'thisSession' type: 'SESSIONIDENTIFIERTYPE'.
	sqFileOop _ interpreterProxy stackValue: 1.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	sqFile _ interpreterProxy arrayValueOf: sqFileOop.
	thisSession _ self sessionIdentifierFrom: (interpreterProxy stackObjectValue: 0).
	(thisSession = (self cCode: 'sqFile->sessionID' inSmalltalk: [-1]))
		ifFalse:
			[^ interpreterProxy primitiveFail]
		ifTrue:
			[descriptor _ self fileDescriptorFrom: sqFileOop.
			(descriptor < 0) ifTrue: [^ interpreterProxy primitiveFail].
			flags _ self cCode: 'fcntl(descriptor, F_GETFL)'.
			retVal _ self cCode: 'fcntl(descriptor, F_SETFL, flags & ~O_NONBLOCK)'.
			interpreterProxy pop: 3; pushInteger: retVal]

! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl 6/13/2005 00:46'!
primitiveSQFileSetNonBlocking
	"Take a struct SQFile from the stack, and call fcntl() to set the file non-blocking I/O."

	| sqFile thisSession sqFileOop descriptor flags retVal |
	self export: true.
	self var: 'sqFile' declareC: 'SQFile *sqFile'.
	self var: 'thisSession' type: 'SESSIONIDENTIFIERTYPE'.
	sqFileOop _ interpreterProxy stackValue: 0.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	sqFile _ interpreterProxy arrayValueOf: sqFileOop.
	"If the session identifier can be obtained, use it as an additional consistency check"
	thisSession _ self getThisSessionIdentifier.
	thisSession ifNotNil:
		[(thisSession = self cCode: 'sqFile->sessionID' inSmalltalk: [-1])
			ifFalse: [^ interpreterProxy primitiveFail]].
	descriptor _ self fileDescriptorFrom: sqFileOop.
	(descriptor < 0) ifTrue: [^ interpreterProxy primitiveFail].
	flags _ self cCode: 'fcntl(descriptor, F_GETFL)'.
	retVal _ self cCode: 'fcntl(descriptor, F_SETFL, flags | O_NONBLOCK)'.
	interpreterProxy pop: 2; pushInteger: retVal

! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl 6/13/2005 00:46'!
primitiveSQFileSetNonBlockingWithSessionIdentifier
	"Take a struct SQFile from the stack, and call fcntl() to set the file non-blocking I/O.
	Use this variant if the session identifier is not available directly in the VM (as may be
	the case if it is not possible to link from this plugin to a variable elsewhere in the VM)."

	| sqFile sqFileOop descriptor flags retVal thisSession |
	self export: true.
	self var: 'sqFile' declareC: 'SQFile *sqFile'.
	self var: 'thisSession' type: 'SESSIONIDENTIFIERTYPE'.
	sqFileOop _ interpreterProxy stackValue: 1.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	sqFile _ interpreterProxy arrayValueOf: sqFileOop.
	thisSession _ self sessionIdentifierFrom: (interpreterProxy stackObjectValue: 0).
	(thisSession = (self cCode: 'sqFile->sessionID' inSmalltalk: [-1]))
		ifFalse:
			[^ interpreterProxy primitiveFail]
		ifTrue:
			[descriptor _ self fileDescriptorFrom: sqFileOop.
			(descriptor < 0) ifTrue: [^ interpreterProxy primitiveFail].
			flags _ self cCode: 'fcntl(descriptor, F_GETFL)'.
			retVal _ self cCode: 'fcntl(descriptor, F_SETFL, flags | O_NONBLOCK)'.
			interpreterProxy pop: 3; pushInteger: retVal]

! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl 6/13/2005 00:46'!
primitiveSQFileSetUnbuffered
	"Take a struct SQFile from the stack, and call setbuf() to set the OS file stream
	(implemented in the C library) for unbuffered I/O. Answers the result of a fflush()
	call, not the result of the setbuf() call (which is type void). This is nearly useless,
	but may at least provide an indicator that we are operating on a valid file stream."

	| sqFileOop sqFile thisSession retVal |
	self export: true.
	self var: 'sqFile' declareC: 'SQFile *sqFile'.
	self var: 'thisSession' type: 'SESSIONIDENTIFIERTYPE'.
	sqFileOop _ interpreterProxy stackValue: 0.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	sqFile _ interpreterProxy arrayValueOf: sqFileOop.
	"If the session identifier can be obtained, use it as an additional consistency check"
	thisSession _ self getThisSessionIdentifier.
	thisSession ifNotNil:
		[(thisSession = self cCode: 'sqFile->sessionID' inSmalltalk: [-1])
			ifFalse: [^ interpreterProxy primitiveFail]].
	retVal _ self cCode: 'fflush(sqFile->file)'.
	self cCode: 'setbuf(sqFile->file, NULL)'.
	interpreterProxy pop: 2; pushInteger: retVal

! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl 6/13/2005 00:46'!
primitiveSQFileSetUnbufferedWithSessionIdentifier
	"Take a struct SQFile from the stack, and call setbuf() to set the OS file stream (implemented in
	the C library) for unbuffered I/O. Answers the result of a fflush() call, not the result of the
	setbuf() call (which is type void). This is nearly useless, but may at least provide an indicator
	that we are operating on a valid file stream. Use this variant if the session identifier is not
	available directly in the VM (as may be the case if it is not possible to link from this plugin
	to a variable elsewhere in the VM)."

	| sqFileOop sqFile retVal thisSession |
	self export: true.
	self var: 'sqFile' declareC: 'SQFile *sqFile'.
	self var: 'thisSession' type: 'SESSIONIDENTIFIERTYPE'.
	sqFileOop _ interpreterProxy stackValue: 1.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	sqFile _ interpreterProxy arrayValueOf: sqFileOop.
	thisSession _ self sessionIdentifierFrom: (interpreterProxy stackObjectValue: 0).
	(thisSession = (self cCode: 'sqFile->sessionID' inSmalltalk: [-1]))
		ifFalse: [^ interpreterProxy primitiveFail]
		ifTrue: [retVal _ self cCode: 'fflush(sqFile->file)'.
				self cCode: 'setbuf(sqFile->file, NULL)'.
				interpreterProxy pop: 3; pushInteger: retVal]

! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl 6/13/2005 00:56'!
primitiveUnixFileClose: anIntegerFileNumber
	"Close a file handle at the close(2) level, using a handle returned by
	#primitiveUnixFileNumber."

	|  handle result |
	self export: true.
	self var: 'handle' type: 'int'.
	self var: 'result' type: 'int'.
	handle _ interpreterProxy stackIntegerValue: 0.
	result _ self cCode: 'close(handle)' inSmalltalk: [-1].
	interpreterProxy pop: 2; pushInteger: result

! !

!UnixOSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl 6/13/2005 00:47'!
primitiveUnixFileNumber
	"Take a struct SQFile from the stack, and answer the value of its Unix file number.
	The Unix file number is not directly useful to Squeak, but may be interesting for
	debugging problems involving failure to close unused file handles."

	| sqFileOop sqFile fileNo |
	self export: true.
	self var: 'sqFile' declareC: 'SQFile *sqFile'.
	self var: 'fileNo' type: 'int'.
	sqFileOop _ interpreterProxy stackValue: 0.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	sqFile _ interpreterProxy arrayValueOf: sqFileOop.
	fileNo _ self cCode: 'fileno(sqFile->file)' inSmalltalk: [-1].
	interpreterProxy pop: 2; pushInteger: fileNo

! !

!UnixOSProcessPlugin methodsFor: 'primitives - file locking' stamp: 'dtl 6/13/2005 00:30'!
primitiveLockFileRegion
	"Take a struct SQFile from the stack, and request a lock on the specified region.
	If the exclusive flag is true, then request an exclusive (F_WRLCK) lock on the
     file. Otherwise, request a shared (F_RDLCK) lock. Any number of Unix processes
     may hold  a read lock (shared lock) on a file region, but only one process may
     hold a write lock (exclusive lock). Answer the result of the call to fcntl().

	If length is zero, then the entire file will be locked, including region extents that
	have not yet been allocated for the file."

	| sqFile thisSession lockStruct exclusive len start sqFileOop fileNo result |
	self export: true.
	self var: 'sqFile' declareC: 'SQFile *sqFile'.
	self var: 'thisSession' type: 'SESSIONIDENTIFIERTYPE'.
	self var: 'lockStruct' declareC: 'struct flock lockStruct'.
	self var: 'fileNo' type: 'int'.
	self var: 'result' type: 'int'.

	"Get the parameters from the stack"
	exclusive _ (interpreterProxy stackValue: 0) == interpreterProxy trueObject.
	len _ interpreterProxy stackIntegerValue: 1.
	start _ interpreterProxy stackIntegerValue: 2.
	sqFileOop _ interpreterProxy stackValue: 3.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	sqFile _ interpreterProxy arrayValueOf: sqFileOop.
	"If the session identifier can be obtained, use it as an additional consistency check"
	thisSession _ self getThisSessionIdentifier.
	thisSession ifNotNil:
		[(thisSession = self cCode: 'sqFile->sessionID' inSmalltalk: [-1])
			ifFalse: [^ interpreterProxy primitiveFail]].
	fileNo _ self cCode: 'fileno(sqFile->file)' inSmalltalk: [-1].

	"Set up the flock structure parameter for fcntl()"
	exclusive
		ifTrue: [self cCode: 'lockStruct.l_type = F_WRLCK']
		ifFalse: [self cCode: 'lockStruct.l_type = F_RDLCK'].
	self cCode: 'lockStruct.l_whence = SEEK_SET'.
	self cCode: 'lockStruct.l_start = start'.
	self cCode: 'lockStruct.l_len = len'.
	self cCode: 'lockStruct.l_pid = 0'.

	"Request the lock and answer the result of the fcntl call"
	result _ self cCode: 'fcntl(fileNo, F_SETLK, &lockStruct)'.
	interpreterProxy pop: 5; pushInteger: result
! !

!UnixOSProcessPlugin methodsFor: 'primitives - file locking' stamp: 'dtl 6/13/2005 00:40'!
primitiveTestLockableFileRegion
	"Take a struct SQFile from the stack, and check for ability to lock the specified region.
	If the exclusive flag is true, then specify an exclusive (F_WRLCK) lock on the
     file. Otherwise, specify a shared (F_RDLCK) lock. Any number of Unix processes
     may hold  a read lock (shared lock) on a file region, but only one process may
     hold a write lock (exclusive lock).

	If length is zero, then the request is for the entire file to be locked, including
	region extents that have not yet been allocated for the file.

	If the fcntl() call fails, answer -1 (the result of the failed call). Otherwise,
	answer an array with the following six fields:
		lockable (true or false)
		l_pid (pid of the process preventing this lock request, or nil)
		l_type (request type F_WRLCK or F_RDLOCK of the process preventing this lock request)
		l_whence (the SEEK_SET, SEEK_CUR, or SEEK_END value of the lock preventing this lock request).
		l_start (offset of the region lock preventing this lock request)
		l_len (length of the region lock preventing this lock request)"

	| sqFile thisSession lockStruct exclusive len start sqFileOop fileNo result resultArray canObtainLock |
	self export: true.
	self var: 'sqFile' declareC: 'SQFile *sqFile'.
	self var: 'thisSession' type: 'SESSIONIDENTIFIERTYPE'.
	self var: 'lockStruct' declareC: 'struct flock lockStruct'.
	self var: 'fileNo' type: 'int'.
	self var: 'result' type: 'int'.

	"Get the parameters from the stack"
	exclusive _ (interpreterProxy stackValue: 0) == interpreterProxy trueObject.
	len _ interpreterProxy stackIntegerValue: 1.
	start _ interpreterProxy stackIntegerValue: 2.
	sqFileOop _ interpreterProxy stackValue: 3.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	sqFile _ interpreterProxy arrayValueOf: sqFileOop.
	"If the session identifier can be obtained, use it as an additional consistency check"
	thisSession _ self getThisSessionIdentifier.
	thisSession ifNotNil:
		[(thisSession = self cCode: 'sqFile->sessionID' inSmalltalk: [-1])
			ifFalse: [^ interpreterProxy primitiveFail]].
	fileNo _ self cCode: 'fileno(sqFile->file)' inSmalltalk: [-1].

	"Set up the flock structure parameter for fcntl()"
	exclusive
		ifTrue: [self cCode: 'lockStruct.l_type = F_WRLCK']
		ifFalse: [self cCode: 'lockStruct.l_type = F_RDLCK'].
	self cCode: 'lockStruct.l_whence = SEEK_SET'.
	self cCode: 'lockStruct.l_start = start'.
	self cCode: 'lockStruct.l_len = len'.
	self cCode: 'lockStruct.l_pid = 0'.

	"Check availability of the lock"
	result _ self cCode: 'fcntl(fileNo, F_GETLK, &lockStruct)'.
	(result == -1)
		ifTrue:
			[interpreterProxy pop: 5; pushInteger: result]
		ifFalse:
			[(self cCode: 'lockStruct.l_type == F_UNLCK')
				ifTrue: [canObtainLock _ interpreterProxy trueObject]
				ifFalse: [canObtainLock _ interpreterProxy falseObject].
			
resultArray _ interpreterProxy
							instantiateClass: (interpreterProxy classArray)
							indexableSize: 6.
			interpreterProxy
				stObject: resultArray
				at: 1
				put: canObtainLock.
			interpreterProxy
				stObject: resultArray
				at: 2
				put: (interpreterProxy integerObjectOf: (self cCode: 'lockStruct.l_pid')).
			interpreterProxy
				stObject: resultArray
				at: 3
				put: (interpreterProxy integerObjectOf: (self cCode: 'lockStruct.l_type')).
			interpreterProxy
				stObject: resultArray
				at: 4
				put: (interpreterProxy integerObjectOf: (self cCode: 'lockStruct.l_whence')).
			interpreterProxy
				stObject: resultArray
				at: 5
				put: (interpreterProxy integerObjectOf: (self cCode: 'lockStruct.l_start')).
			interpreterProxy
				stObject: resultArray
				at: 6
				put: (interpreterProxy integerObjectOf: (self cCode: 'lockStruct.l_len')).
			interpreterProxy pop: 5 thenPush: resultArray]
! !

!UnixOSProcessPlugin methodsFor: 'primitives - file locking' stamp: 'dtl 6/13/2005 00:41'!
primitiveUnlockFileRegion
	"Take a struct SQFile from the stack, and unlock the specified region.
	Answer the result of the call to fcntl(). If the region is in the file lock cache,
	remove it, but otherwise ignore the cache. The cache supports Win32 semantics
	within a single Squeak image, but not across separate images, therefore the
	unlock should be attempted regardless of whether this image thinks that the
	region has previously been locked. Answer the result of the call to fcntl()."

	| sqFile thisSession lockStruct len start sqFileOop fileNo result |
	self export: true.
	self var: 'sqFile' declareC: 'SQFile *sqFile'.
	self var: 'thisSession' type: 'SESSIONIDENTIFIERTYPE'.
	self var: 'lockStruct' declareC: 'struct flock lockStruct'.
	self var: 'fileNo' type: 'int'.
	self var: 'result' type: 'int'.

	"Get the parameters from the stack"
	len _ interpreterProxy stackIntegerValue: 0.
	start _ interpreterProxy stackIntegerValue: 1.
	sqFileOop _ interpreterProxy stackValue: 2.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	sqFile _ interpreterProxy arrayValueOf: sqFileOop.
	"If the session identifier can be obtained, use it as an additional consistency check"
	thisSession _ self getThisSessionIdentifier.
	thisSession ifNotNil:
		[(thisSession = self cCode: 'sqFile->sessionID' inSmalltalk: [-1])
			ifFalse: [^ interpreterProxy primitiveFail]].
	fileNo _ self cCode: 'fileno(sqFile->file)' inSmalltalk: [-1].

	"Set up the flock structure parameter for fcntl()"
	self cCode: 'lockStruct.l_type = F_UNLCK'.
	self cCode: 'lockStruct.l_whence = SEEK_SET'.
	self cCode: 'lockStruct.l_start = start'.
	self cCode: 'lockStruct.l_len = len'.
	self cCode: 'lockStruct.l_pid = 0'.

	"Request the lock and answer the result of the fcntl call"
	result _ self cCode: 'fcntl(fileNo, F_SETLK, &lockStruct)'.
	interpreterProxy pop: 4; pushInteger: result
! !

!UnixOSProcessPlugin methodsFor: 'primitives - fork and exec' stamp: 'dtl 7/12/2005 07:05'!
forkAndExecInDirectory: useSignalHandler
	"Fork a child OS process, and do an exec in the child. The parent continues on in
	Smalltalk, and this method answers the pid of the child which was created. If
	useSignalHandler is true, set the signal handler for SIGCHLD. Otherwise, assume
	that death of child events are handled through some other mechanism.

	In this implementation, memory for the argument and environment arrays is allocated
	in the image prior to calling this primitive. This allows us to avoid invoking the
	garbage collector in this primitive (thereby moving the locations of environment
	and argument memory), but comes at the cost of twiddling C pointers here in the
	primitive. An alternative to this whole mess is just to malloc the environment and
	argument vectors, but I think it is a good idea to avoid malloc as much as possible
	so as not to limit future ObjectMemory implementations.

	This primitive replaces #primitiveForkAndExec from earlier versions of the plugin.
	The new name permits backward compatibility for an image running on a VM
	which does not yet have the updated plugin. This implementation uses a different
	argument format on the stack, and differs functionally in that the child now closes
	all file descriptors (including sockets) not required (that is, everything except stdin,
	stdout, and stderr on descriptors 0, 1 and 2). This eliminates some flakey behavior
	in child processes connected to Squeak by pipes, which failed to exit at expected times
	due to the old file descriptors remaining open. This is also cleaner in that garbage
	descriptors are not left hanging around the the child.

	On entry, the stack contains:
		0: workingDir, a null terminated string specifying the working directory to use, or nil.
		1: envOffsets, an array of integers for calculating environment string address offsets.
		2: envVecBuffer, a String buffer containing environment strings arranged to look like char **.
		3: argOffsets, an array of integers for calculating argument string address offsets.
		4: argVecBuffer, a String buffer containing argument strings arranged to look like char **.
		5: stdErr, a ByteArray handle.
		6: stdOut, a ByteArray handle.
		7: stdIn, a ByteArray handle.
		8: executableFile, a null terminated string with the name of the file to execute.
		9: the sender"

	| progNamePtr argsPtr envPtr intervalTimer saveIntervalTimer envVec pid
	envOffsets envVecBuffer argOffsets argVecBuffer stdErr stdOut stdIn executableFile idx
	handleCount envCount argCount envOffsetPtr argOffsetPtr workingDir pwdPtr pwd env args |
	self var: 'progNamePtr' declareC: 'char *progNamePtr'.
	self var: 'argsPtr' declareC: 'char *argsPtr'.
	self var: 'envPtr' declareC: 'char *envPtr'.
	self var: 'args' declareC: 'char **args'.
	self var: 'env' declareC: 'char **env'.
	self var: 'pwdPtr' declareC: 'char *pwdPtr'.
	self var: 'intervalTimer' type: 'struct itimerval'.
	self var: 'saveIntervalTimer' type: 'struct itimerval'.
	self var: 'envVec' declareC: 'extern char **envVec'.
	self var: 'pid' type: 'pid_t'.
	self var: 'envOffsetPtr' declareC: 'sqInt *envOffsetPtr'.
	self var: 'argOffsetPtr' declareC: 'sqInt *argOffsetPtr'.

	useSignalHandler ifTrue: [self setSigChldHandler].

	"Do not fork child if running in secure mode"
	(self sandboxSecurity == 1) ifTrue: [interpreterProxy pop: 10; pushInteger: -1] ifFalse: [

	"Turn off the interval timer. If this is not done, then the program which we exec in
	the child process will receive a timer interrupt, and will not know how to handle it."
	self cCode: 'intervalTimer.it_interval.tv_sec = 0'.
	self cCode: 'intervalTimer.it_interval.tv_usec = 0'.
	self cCode: 'intervalTimer.it_value.tv_sec = 0'.
	self cCode: 'intervalTimer.it_value.tv_usec = 0'.
	self cCode: 'setitimer (ITIMER_REAL, &intervalTimer, &saveIntervalTimer)'.

	((pid _ self cCode: 'vfork()') = 0)
		ifFalse:
			[ "Normal return to Smalltalk - this is the old parent process."
			"Enable the timer again before resuming Smalltalk."
			self cCode: 'setitimer (ITIMER_REAL, &saveIntervalTimer, 0L)'.
			interpreterProxy pop: 10; pushInteger: pid 
			"Pop 9 arguments plus receiver, push pid."]
		ifTrue:
			[ "This is the new child process"
			workingDir _ interpreterProxy stackObjectValue: 0.
			envOffsets _ interpreterProxy stackObjectValue: 1.
			envVecBuffer _ interpreterProxy stackObjectValue: 2.
			argOffsets _ interpreterProxy stackObjectValue: 3.
			argVecBuffer _ interpreterProxy stackObjectValue: 4.
			stdErr _ interpreterProxy stackObjectValue: 5.
			stdOut _ interpreterProxy stackObjectValue: 6.
			stdIn _ interpreterProxy stackObjectValue: 7.
			executableFile _ interpreterProxy stackObjectValue: 8.

			"If a new working directory has been specified, try to chdir() to it."
			(workingDir ~= interpreterProxy nilObject)
				ifTrue:
					[pwdPtr _ interpreterProxy firstIndexableField: workingDir.
					(pwdPtr == 0)
						ifTrue:
							[self cCode: 'fprintf(stderr, "bad workingDir parameter\n")'.
							self cCode: '_exit(-1)']
						ifFalse:
							[pwd _ self cCode: 'chdir(pwdPtr)' inSmalltalk: [0].
							(pwd ~= 0)
								ifTrue:
									[self cCode: 'perror("chdir")'.
									self cCode: '_exit(-1)']]].

			"Pointer to name of the file to execute."
			progNamePtr _ interpreterProxy arrayValueOf: executableFile.

			"Dup the file handles to attach the new child process to the right streams
			on descriptors 0, 1 and 2."
			(stdErr = interpreterProxy nilObject) ifFalse: [self dupToStdErr: stdErr].
			(stdOut = interpreterProxy nilObject) ifFalse: [self dupToStdOut: stdOut].
			(stdIn = interpreterProxy nilObject) ifFalse: [self dupToStdIn: stdIn].

			"Close all file handles not needed by the child. This is done by brute force
			because the image has no way of being aware of all open file descriptors
			(some are opened by the VM without the knowledge of the image)."
			idx _ 3.	"First Unix file descriptor after stdin, stdout, stderr."
			handleCount _ self descriptorTableSize.
			[idx < handleCount]
				whileTrue:
					[self cCode: 'close(idx)'.
					idx _ idx + 1].

			"Set up the environment, a C array of pointers to C strings, where each string
			is of the form 'KEY=value'. Refer to previously allocated object memory rather
			than using a call to malloc()."
			envVecBuffer = (interpreterProxy nilObject)
				ifTrue:
					[env _ envVec]
				ifFalse:
					[envCount _ interpreterProxy stSizeOf: envOffsets.
					envPtr _ interpreterProxy arrayValueOf: envVecBuffer.
					envOffsetPtr _ interpreterProxy firstIndexableField: envOffsets.
					self
						fixPointersInArrayOfStrings: envPtr
						withOffsets: envOffsetPtr
						count: envCount.
					env _ self cCoerce: envPtr to: 'char **'].

			"Set up the argument array. Allocate space for a C array of pointers to strings.
			Reer to previously allocated object memory rather than using a call to malloc()."
			argCount _ interpreterProxy stSizeOf: argOffsets.
			argsPtr _ interpreterProxy arrayValueOf: argVecBuffer.
			argOffsetPtr _ interpreterProxy firstIndexableField: argOffsets.
			self
				fixPointersInArrayOfStrings: argsPtr
				withOffsets: argOffsetPtr
				count: argCount.
			args _ self cCoerce: argsPtr to: 'char **'.

			"Clean things up before clobbering the running image."
			"Note: If any file descriptors, signal handlers, or other references to external
			resources need to be cleaned up, do it here."
			self restoreDefaultSignalHandlers.

			"Do the exec to overlay the new program on this child address space. This is 
			the end of the world as we know it."
			(self cCode: 'execve(progNamePtr, args, env)' ) = -1
				ifTrue:
					[self cCode: 'perror(progNamePtr)'; cCode: '_exit(-1)' "execve() error, child exits"]
				ifFalse:
					[self cCode: '/* Can''t get here from there */']]]
! !

!UnixOSProcessPlugin methodsFor: 'primitives - fork and exec' stamp: 'dtl 6/12/2005 15:23'!
forkSqueak: useSignalHandler
	"Fork a child process, and continue running squeak in the child process. If displayFlag
	is true, open a new X display for the child, otherwise the child is a headless Squeak.
	Answer the result of the fork() call, either the child pid or zero.

	After calling fork(), two OS processes exist, one of which is the child of the other. On
	systems which implement copy-on-write memory management, and which support the
	fork() system call, both processes will be running Smalltalk images, and will be sharing
	the same memory space. In the original OS process, the resulting value of pid is the
	process id of the child process (a non-zero integer). In the child process, the value of
	pid is zero.

	The child recreates sufficient external resources to continue running. This is done by
	attaching to a new X session. The child is otherwise a copy of the parent process, and
	will continue executing the Smalltalk image at the same point as its parent. The return
	value of this primitive may be used by the two running Smalltalk images to determine
	which is the parent and which is the child.

	The child should not depend on using existing connections to external resources. For
	example, the child may lose its connections to stdin, stdout, and stderr after its parent
	exits.

	The new child image does not start itself from the image in the file system; rather it is
	a clone of the parent image as it existed at the time of primitiveForkSqueak. For this
	reason, the parent and child should agree in advance as to whom is allowed to save the
	image to the file system, otherwise one Smalltalk may overwrite the image of the other.

	This is a simple call to fork(), rather than the more common idiom of vfork() followed
	by exec(). The vfork() call cannot be used here because it is designed to be followed by
	an exec(), and its semantics require the parent process to wait for the child to exit. See
	the BSD programmers documentation for details."

	| pid intervalTimer saveIntervalTimer |
	self export: true.
	self returnTypeC: 'pid_t'.
	self var: 'pid' type: 'pid_t'.
	self var: 'intervalTimer' type: 'struct itimerval'.
	self var: 'saveIntervalTimer' type: 'struct itimerval'.

	"Turn off the interval timer. If this is not done, then the program which we exec in
	the child process will receive a timer interrupt, and will not know how to handle it."
	self cCode: 'intervalTimer.it_interval.tv_sec = 0'.
	self cCode: 'intervalTimer.it_interval.tv_usec = 0'.
	self cCode: 'intervalTimer.it_value.tv_sec = 0'.
	self cCode: 'intervalTimer.it_value.tv_usec = 0'.
	self cCode: 'setitimer (ITIMER_REAL, &intervalTimer, &saveIntervalTimer)'.
	useSignalHandler ifTrue: [self setSigChldHandler].
	pid _ self fork.

	"Enable the timer again before resuming Smalltalk."
	self cCode: 'setitimer (ITIMER_REAL, &saveIntervalTimer, 0L)'.
	^ pid
! !

!UnixOSProcessPlugin methodsFor: 'primitives - fork and exec' stamp: 'dtl 1/25/2004 19:31'!
primitiveForkExec
	"Fork a child OS process, and do an exec in the child. The parent continues on in
	Smalltalk, and this method answers the pid of the child which was created.

	On entry, the stack contains:
		0: workingDir, a null terminated string specifying the working directory to use, or nil.
		1: envOffsets, an array of integers for calculating environment string address offsets.
		2: envVecBuffer, a String buffer containing environment strings arranged to look like char **.
		3: argOffsets, an array of integers for calculating argument string address offsets.
		4: argVecBuffer, a String buffer containing argument strings arranged to look like char **.
		5: stdErr, a ByteArray handle.
		6: stdOut, a ByteArray handle.
		7: stdIn, a ByteArray handle.
		8: executableFile, a null terminated string with the name of the file to execute.
		9: the sender"

	self export: true.
	^ self forkAndExecInDirectory: false
! !

!UnixOSProcessPlugin methodsFor: 'primitives - fork and exec' stamp: 'dtl 6/12/2005 15:24'!
primitiveForkSqueak
	"Fork a child process, and continue running squeak in the child process. Leave the
	X session connected to the parent process, but close its file descriptor for the child
	process. Open a new X session for the child.

	The child should not depend on using existing connections to external resources. For
	example, the child may lose its connections to stdin, stdout, and stderr after its parent
	exits."

	| pid |
	self export: true.
	self var: 'pid' type: 'pid_t'.

	"Do not fork child if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 1; pushInteger: -1]
		ifFalse:
			[pid _ self forkSqueak: true.
			interpreterProxy pop: 1; pushInteger: pid]
! !

!UnixOSProcessPlugin methodsFor: 'primitives - fork and exec' stamp: 'dtl 6/12/2005 15:24'!
primitiveForkSqueakWithoutSigHandler
	"Fork a child process, and continue running squeak in the child process. Leave the
	X session connected to the parent process, but close its file descriptor for the child
	process. Open a new X session for the child.

	The child should not depend on using existing connections to external resources. For
	example, the child may lose its connections to stdin, stdout, and stderr after its parent
	exits."

	| pid |
	self export: true.
	self var: 'pid' type: 'pid_t'.

	"Do not fork child if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 1; pushInteger: -1]
		ifFalse:
			[pid _ self forkSqueak: false.
			interpreterProxy pop: 1; pushInteger: pid]
! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 2/21/2002 15:47'!
primitiveArgumentAt
	"Answer a string containing the OS process argument at index (an Integer) in the
	argument list."

	| index argCnt argVec sPtr s |
	self export: true.
	self var: 'argCnt' declareC: 'extern int argCnt'.
	self var: 'argVec' declareC: 'extern char **argVec'.
	self var: 'sPtr' declareC: 'char *sPtr'.
	index _ interpreterProxy stackIntegerValue: 0.
	((index > argCnt) | (index < 1))
	ifTrue:
		[interpreterProxy pop: 2; push: interpreterProxy nilObject]
	ifFalse:
		[sPtr _ self cCode: 'argVec[index - 1]' inSmalltalk: [''].
		s _ self stringFromCString: sPtr.
		interpreterProxy pop: 2; push: s]

! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 2/21/2002 15:47'!
primitiveChdir
	"Call chdir(2) to change current working directory to the specified path string. Answer
	nil for success, or errno on failure."

	| path errno |
	self export: true.
	self var: 'path' declareC: 'char *path'.
	self var: 'errno' declareC: 'extern int errno'.

	path _ self transientCStringFromString: (interpreterProxy stackObjectValue: 0).
	((self cCode: 'chdir(path)' inSmalltalk: [0]) == 0)
		ifTrue: [interpreterProxy pop: 2; push: interpreterProxy nilObject]
		ifFalse: [interpreterProxy pop: 2; pushInteger: errno].

! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 2/21/2002 15:47'!
primitiveEnvironmentAt
	"Answer a string containing the OS process environment string at index (an Integer)
	in the environment list."

	| index envVec p envCnt sPtr s |
	self export: true.
	self var: 'p' declareC: 'char **p'.
	self var: 'sPtr' declareC: 'char *sPtr'.
	self var: 'envVec' declareC: 'extern char **envVec'.
	"Count number of environment variables."
	p _ envVec.
	envCnt _ 0.
	self cCode: 'while (*p++) envCnt++' inSmalltalk: [envCnt _ 0].
	index _ interpreterProxy stackIntegerValue: 0.
	((index > envCnt) | (index < 1))
	ifTrue:
		[interpreterProxy pop: 2; push: interpreterProxy nilObject]
	ifFalse:
		[sPtr _ self cCode: 'envVec[index - 1]' inSmalltalk: [''].
		s _ self stringFromCString: sPtr.
		interpreterProxy pop: 2; push: s]
! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 2/21/2002 15:15'!
primitiveEnvironmentAtSymbol
	"Answer the value of an environment variable keyed by a Symbol."

	| keyCStringPtr getenvResult s |
	self export: true.
	self var: 'keyCStringPtr' declareC: 'char *keyCStringPtr'.
	self var: 'getenvResult' declareC: 'char * getenvResult'.

	keyCStringPtr _ self transientCStringFromString: (interpreterProxy stackObjectValue: 0).
	getenvResult _ self cCode: 'getenv(keyCStringPtr)' inSmalltalk: [0].
	(getenvResult = 0)
		ifTrue:
			[^ interpreterProxy primitiveFail]
		ifFalse:
			[s _ self stringFromCString: getenvResult.
			interpreterProxy pop: 2; push: s]

! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 2/21/2002 15:47'!
primitiveErrorMessageAt
	"Answer a string describing an error message"

	| index errMessage p |
	self export: true.
	self var: 'p' declareC: 'char *p'.

	index _ interpreterProxy stackIntegerValue: 0.
	p _ self cCoerce: (self cCode: 'strerror(index)' inSmalltalk: ['']) to: 'char *'.
	errMessage _ self stringFromCString: p.
	interpreterProxy pop: 2; push: errMessage
! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 2/21/2002 15:16'!
primitiveGetCurrentWorkingDirectory
	"Answer a string containing the current working directory."

	| cwd cwdString bufferSize incrementBy maxSize buffer |
	self export: true.
	self var: 'buffer' declareC: 'char *buffer'.
	self var: 'cwd' declareC: 'char *cwd'.

	bufferSize _ 100.
	incrementBy _ 100.
	maxSize _ 5000.

	[cwdString _ interpreterProxy
		instantiateClass: interpreterProxy classString
		indexableSize: bufferSize.
	buffer _ interpreterProxy arrayValueOf: cwdString.
	cwd _ self cCode: 'getcwd(buffer, bufferSize)' inSmalltalk: [''].
	((cwd == 0)
		and: [bufferSize < maxSize])]
			whileTrue:
				[bufferSize _ bufferSize + incrementBy].
	(cwd == 0)
		ifTrue:
			[^ interpreterProxy primitiveFail]
		ifFalse:
			[cwdString _ self stringFromCString: cwd.
			interpreterProxy pop: 1; push: cwdString]
! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 6/13/2005 00:12'!
primitiveGetEGid
	"Answer the effective group ID of my OS process"

	| eGid |
	self export: true.
	self var: 'eGid' type: 'gid_t'.
	eGid := self cCode: 'getegid()'.
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: eGid! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 6/13/2005 00:14'!
primitiveGetEUid
	"Answer the effective user ID of my OS process"

	| eUid |
	self export: true.
	self var: 'eUid' type: 'uid_t'.
	eUid := self cCode: 'geteuid()'.
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: eUid! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 6/13/2005 00:13'!
primitiveGetGid
	"Answer the group ID of my OS process"

	| gid |
	self export: true.
	self var: 'gid' type: 'gid_t'.
	gid := self cCode: 'getgid()'.
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: gid! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 6/13/2005 00:13'!
primitiveGetPid
	"Answer the process ID of my OS process"

	| pid |
	self export: true.
	self var: 'pid' type: 'pid_t'.
	pid := self cCode: 'getpid()'.
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: pid! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 6/13/2005 00:15'!
primitiveGetPPid
	"Answer the process ID of the parent process of my OS process"

	| ppid |
	self export: true.
	self var: 'ppid' type: 'pid_t'.
	ppid := self cCode: 'getppid()'.
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: ppid! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 6/13/2005 00:14'!
primitiveGetUid
	"Answer the user ID of my OS process"

	| uid |
	self export: true.
	self var: 'uid' type: 'uid_t'.
	uid := self cCode: 'getuid()'.
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: uid! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 10/8/2005 09:43'!
primitiveNice
	"Change the scheduling priority of this process by the given nice increment.
	A positive increment decreases the priority. Only the superuser can specify
	a negative value (to increase the priority). See man(2) nice.

	Different versions of Unix are inconsistent in their return values. The only
	reliable test for success is to clear errno prior to the call, and test its value
	if the result of nice() is -1."

	| niceIncrement errno result |
	self export: true.
	self var: 'errno' type: 'extern int'.
	self var: 'niceIncrement' type: 'int'.
	niceIncrement _ interpreterProxy stackIntegerValue: 0.
	errno _ 0.
	result _ self nice: niceIncrement.
	(result == -1) ifTrue:
		["sys call may have failed, test errno to be sure"
		(errno == 0) ifFalse:
			[^ interpreterProxy primitiveFail]].
	interpreterProxy pop: 2; pushInteger: result.
! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 2/21/2002 15:19'!
primitivePutEnv
	"Set an environment variable using a string of the form 'KEY=value'. This
	implementation allocates a C string using malloc to allocate from the C heap
	(using cStringFromString rather than transientCStringFromString). This
	is necessary because the C runtime library does not make a copy of the
	string into separately allocated environment memory."

	| cStringPtr keyValueString |
	self export: true.
	self var: 'cStringPtr' declareC: 'char *cStringPtr'.
	keyValueString _ interpreterProxy stackObjectValue: 0.
	cStringPtr _ self cStringFromString: keyValueString.
	((self cCode: 'putenv(cStringPtr)') = 0)	"Set environment variable."
		ifTrue: [interpreterProxy pop: 2; push: keyValueString]
		ifFalse: [^ interpreterProxy primitiveFail]
! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 7/12/2005 22:59'!
primitiveReapChildProcess
	"Clean up after the death of a child, and answer an array with the pid and
	the exit status of the child process. Answer nil if the pidToHandle does not exist."

	| pidToHandle pidResult exitStatus resultArray arrayPtr |
	self export: true.
	self var: 'exitStatus' type: 'int'.
	self var: 'pidResult' type: 'pid_t'.
	self var: 'arrayPtr' declareC: 'sqInt *arrayPtr'.
	exitStatus _ 0.	"Force C code translator to declare the variable"
	pidToHandle _ interpreterProxy stackIntegerValue: 0.
	pidResult _ self cCode: 'waitpid ( pidToHandle, &exitStatus, WNOHANG )'
					inSmalltalk: [ exitStatus _ -1 ].
	pidResult <= 0
		ifTrue: [ interpreterProxy pop: 2; push: interpreterProxy nilObject ]
		ifFalse: [ "Answer an array with pid and result status "
			resultArray _ interpreterProxy
							instantiateClass: interpreterProxy classArray
							indexableSize: 2.
			arrayPtr _ interpreterProxy firstIndexableField: resultArray.
			self cCode: 'arrayPtr[0] = integerObjectOf(pidResult)'
				inSmalltalk: [resultArray at: 1 put: pidResult].
			self cCode: 'arrayPtr[1] = integerObjectOf(exitStatus)'
				inSmalltalk: [resultArray at: 2 put: exitStatus].
			interpreterProxy pop: 2; push: resultArray ]
! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 2/21/2002 15:48'!
primitiveUnsetEnv
	"Unset an environment variable."

	"FIXME: unsetenv() is not portable. For Solaris or any other system which does not
	support unsetenv(), just comment it out in this method and rebuild the plugin."

	| cStringPtr keyString |
	self export: true.
	self var: 'cStringPtr' declareC: 'char *cStringPtr'.
	keyString _ interpreterProxy stackObjectValue: 0.
	cStringPtr _ self transientCStringFromString: keyString.
	"For Solaris, comment out the following line"
	self cCode: 'unsetenv(cStringPtr)'.	"Unset environment variable."
	interpreterProxy pop: 1
! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl 6/13/2005 00:50'!
primitiveCanReceiveSignals: anIntegerPid
	"Send a null signal to the OS process identified by anIntegerPid. Answer false for
	a bad parameter on the stack (the common case is for anIntegerPid equal to nil,
	for which case we should answer false). Answer true if the process exists and can
	receive signals from this process, otherwise false. This test is useful for determining
	if a child process still exists following a Squeak image restart."

	| result pidToSignal |
	self export: true.
	self var: 'pidToSignal' type: 'pid_t'.
	self var: 'result' type: 'int'.
	(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
		ifTrue: 
			[pidToSignal _ interpreterProxy stackIntegerValue: 0.
			result _ self cCode: 'kill(pidToSignal, 0)' inSmalltalk: [-1].
			interpreterProxy pop: 2.
			(result == 0)
				ifTrue: [interpreterProxy push: interpreterProxy trueObject]
				ifFalse: [interpreterProxy push: interpreterProxy falseObject]]
		ifFalse: [interpreterProxy pop: 2; push: interpreterProxy falseObject]! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl 6/13/2005 00:51'!
primitiveSendSigabrtTo: anIntegerPid 
	"Send SIGABRT (abort) to the OS process identified by anIntegerPid. Use an explicit
	check for isIntegerObject so we can return -1 on error (the stackIntegerValue: method
	answers 1 on error, and 1 is a valid pid number)."

	| result pidToSignal |
	self export: true.
	self var: 'pidToSignal' type: 'pid_t'.
	self var: 'result' type: 'int'.

	"Do not allow signal sending if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; pushInteger: -1]
		ifFalse:
			[(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
				ifTrue: 
					[pidToSignal _ interpreterProxy stackIntegerValue: 0.
					result _ self cCode: 'kill(pidToSignal, SIGABRT)' inSmalltalk: [-1].
					interpreterProxy pop: 2; pushInteger: result]
				ifFalse: [interpreterProxy pop: 2; pushInteger: -1]]! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl 6/13/2005 00:51'!
primitiveSendSigalrmTo: anIntegerPid 
	"Send SIGALRM (alarm clock) to the OS process identified by anIntegerPid. Use an explicit
	check for isIntegerObject so we can return -1 on error (the stackIntegerValue: method
	answers 1 on error, and 1 is a valid pid number)."

	| result pidToSignal |
	self export: true.
	self var: 'pidToSignal' type: 'pid_t'.
	self var: 'result' type: 'int'.

	"Do not allow signal sending if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; pushInteger: -1]
		ifFalse:
			[(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
				ifTrue: 
					[pidToSignal _ interpreterProxy stackIntegerValue: 0.
					result _ self cCode: 'kill(pidToSignal, SIGALRM)' inSmalltalk: [-1].
					interpreterProxy pop: 2; pushInteger: result]
				ifFalse: [interpreterProxy pop: 2; pushInteger: -1]]! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl 6/13/2005 00:51'!
primitiveSendSigchldTo: anIntegerPid
	"Send SIGCHLD (child status has changed, usually death of child) to the OS process
	identified by anIntegerPid. Use an explicit check for isIntegerObject so we can
	return -1 on error (the stackIntegerValue: method answers 1 on error, and 1 is a
	valid pid number)."

	| result pidToSignal |
	self export: true.
	self var: 'pidToSignal' type: 'pid_t'.
	self var: 'result' type: 'int'.

	"Do not allow signal sending if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; pushInteger: -1]
		ifFalse:
			[(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
				ifTrue: 
					[pidToSignal _ interpreterProxy stackIntegerValue: 0.
					result _ self cCode: 'kill(pidToSignal, SIGCHLD)' inSmalltalk: [-1].
					interpreterProxy pop: 2; pushInteger: result]
				ifFalse: [interpreterProxy pop: 2; pushInteger: -1]]! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl 6/13/2005 00:51'!
primitiveSendSigcontTo: anIntegerPid
	"Send SIGCONT (continue) to the OS process identified by anIntegerPid. Use an explicit
	check for isIntegerObject so we can return -1 on error (the stackIntegerValue: method
	answers 1 on error, and 1 is a valid pid number)."

	| result pidToSignal |
	self export: true.
	self var: 'pidToSignal' type: 'pid_t'.
	self var: 'result' type: 'int'.

	"Do not allow signal sending if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; pushInteger: -1]
		ifFalse:
			[(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
				ifTrue: 
					[pidToSignal _ interpreterProxy stackIntegerValue: 0.
					result _ self cCode: 'kill(pidToSignal, SIGCONT)' inSmalltalk: [-1].
					interpreterProxy pop: 2; pushInteger: result]
				ifFalse: [interpreterProxy pop: 2; pushInteger: -1]]! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl 6/13/2005 00:51'!
primitiveSendSighupTo: anIntegerPid
	"Send SIGHUP (hangup) to the OS process identified by anIntegerPid. Use an explicit
	check for isIntegerObject so we can return -1 on error (the stackIntegerValue: method
	answers 1 on error, and 1 is a valid pid number)."

	| result pidToSignal |
	self export: true.
	self var: 'pidToSignal' type: 'pid_t'.
	self var: 'result' type: 'int'.

	"Do not allow signal sending if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; pushInteger: -1]
		ifFalse:
			[(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
				ifTrue: 
					[pidToSignal _ interpreterProxy stackIntegerValue: 0.
					result _ self cCode: 'kill(pidToSignal, SIGHUP)' inSmalltalk: [-1].
					interpreterProxy pop: 2; pushInteger: result]
				ifFalse: [interpreterProxy pop: 2; pushInteger: -1]]! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl 6/13/2005 00:51'!
primitiveSendSigintTo: anIntegerPid
	"Send SIGINT (interrupt) to the OS process identified by anIntegerPid. Use an explicit
	check for isIntegerObject so we can return -1 on error (the stackIntegerValue: method
	answers 1 on error, and 1 is a valid pid number)."

	| result pidToSignal |
	self export: true.
	self var: 'pidToSignal' type: 'pid_t'.
	self var: 'result' type: 'int'.

	"Do not allow signal sending if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; pushInteger: -1]
		ifFalse:
			[(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
				ifTrue: 
					[pidToSignal _ interpreterProxy stackIntegerValue: 0.
					result _ self cCode: 'kill(pidToSignal, SIGINT)' inSmalltalk: [-1].
					interpreterProxy pop: 2; pushInteger: result]
				ifFalse: [interpreterProxy pop: 2; pushInteger: -1]]! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl 6/13/2005 00:51'!
primitiveSendSigkillTo: anIntegerPid
	"Send SIGKILL (kill, unblockable) to the OS process identified by anIntegerPid. Use an explicit
	check for isIntegerObject so we can return -1 on error (the stackIntegerValue: method
	answers 1 on error, and 1 is a valid pid number)."

	| result pidToSignal |
	self export: true.
	self var: 'pidToSignal' type: 'pid_t'.
	self var: 'result' type: 'int'.

	"Do not allow signal sending if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; pushInteger: -1]
		ifFalse:
			[(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
				ifTrue: 
					[pidToSignal _ interpreterProxy stackIntegerValue: 0.
					result _ self cCode: 'kill(pidToSignal, SIGKILL)' inSmalltalk: [-1].
					interpreterProxy pop: 2; pushInteger: result]
				ifFalse: [interpreterProxy pop: 2; pushInteger: -1]]! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl 6/13/2005 00:51'!
primitiveSendSigpipeTo: anIntegerPid
	"Send SIGPIPE (broken pipe) to the OS process identified by anIntegerPid. Use an explicit
	check for isIntegerObject so we can return -1 on error (the stackIntegerValue: method
	answers 1 on error, and 1 is a valid pid number)."

	| result pidToSignal |
	self export: true.
	self var: 'pidToSignal' type: 'pid_t'.
	self var: 'result' type: 'int'.

	"Do not allow signal sending if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; pushInteger: -1]
		ifFalse:
			[(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
				ifTrue: 
					[pidToSignal _ interpreterProxy stackIntegerValue: 0.
					result _ self cCode: 'kill(pidToSignal, SIGPIPE)' inSmalltalk: [-1].
					interpreterProxy pop: 2; pushInteger: result]
				ifFalse: [interpreterProxy pop: 2; pushInteger: -1]]! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl 6/13/2005 00:51'!
primitiveSendSigquitTo: anIntegerPid
	"Send SIGQUIT (quit) to the OS process identified by anIntegerPid. Use an explicit
	check for isIntegerObject so we can return -1 on error (the stackIntegerValue: method
	answers 1 on error, and 1 is a valid pid number)."

	| result pidToSignal |
	self export: true.
	self var: 'pidToSignal' type: 'pid_t'.
	self var: 'result' type: 'int'.

	"Do not allow signal sending if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; pushInteger: -1]
		ifFalse:
			[(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
				ifTrue: 
					[pidToSignal _ interpreterProxy stackIntegerValue: 0.
					result _ self cCode: 'kill(pidToSignal, SIGQUIT)' inSmalltalk: [-1].
					interpreterProxy pop: 2; pushInteger: result]
				ifFalse: [interpreterProxy pop: 2; pushInteger: -1]]! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl 6/13/2005 00:51'!
primitiveSendSigstopTo: anIntegerPid
	"Send SIGSTOP (stop, unblockable) to the OS process identified by anIntegerPid. Use an explicit
	check for isIntegerObject so we can return -1 on error (the stackIntegerValue: method
	answers 1 on error, and 1 is a valid pid number)."

	| result pidToSignal |
	self export: true.
	self var: 'pidToSignal' type: 'pid_t'.
	self var: 'result' type: 'int'.

	"Do not allow signal sending if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; pushInteger: -1]
		ifFalse:
			[(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
				ifTrue: 
					[pidToSignal _ interpreterProxy stackIntegerValue: 0.
					result _ self cCode: 'kill(pidToSignal, SIGSTOP)' inSmalltalk: [-1].
					interpreterProxy pop: 2; pushInteger: result]
				ifFalse: [interpreterProxy pop: 2; pushInteger: -1]]! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl 6/13/2005 00:52'!
primitiveSendSigtermTo: anIntegerPid
	"Send SIGTERM (termination) to the OS process identified by anIntegerPid. Use an explicit
	check for isIntegerObject so we can return -1 on error (the stackIntegerValue: method
	answers 1 on error, and 1 is a valid pid number)."

	| result pidToSignal |
	self export: true.
	self var: 'pidToSignal' type: 'pid_t'.
	self var: 'result' type: 'int'.

	"Do not allow signal sending if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; pushInteger: -1]
		ifFalse:
			[(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
				ifTrue: 
					[pidToSignal _ interpreterProxy stackIntegerValue: 0.
					result _ self cCode: 'kill(pidToSignal, SIGTERM)' inSmalltalk: [-1].
					interpreterProxy pop: 2; pushInteger: result]
				ifFalse: [interpreterProxy pop: 2; pushInteger: -1]]! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl 6/13/2005 00:52'!
primitiveSendSigusr1To: anIntegerPid
	"Send SIGUSR1 (User-defined signal 1) to the OS process identified by anIntegerPid. Use
	an explicit check for isIntegerObject so we can return -1 on error (the stackIntegerValue:
	method answers 1 on error, and 1 is a valid pid number)."

	| result pidToSignal |
	self export: true.
	self var: 'pidToSignal' type: 'pid_t'.
	self var: 'result' type: 'int'.

	"Do not allow signal sending if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; pushInteger: -1]
		ifFalse:
			[(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
				ifTrue: 
					[pidToSignal _ interpreterProxy stackIntegerValue: 0.
					result _ self cCode: 'kill(pidToSignal, SIGUSR1)' inSmalltalk: [-1].
					interpreterProxy pop: 2; pushInteger: result]
				ifFalse: [interpreterProxy pop: 2; pushInteger: -1]]! !

!UnixOSProcessPlugin methodsFor: 'primitives - OS signals' stamp: 'dtl 6/13/2005 00:52'!
primitiveSendSigusr2To: anIntegerPid
	"Send SIGUSR2 (User-defined signal 2) to the OS process identified by anIntegerPid. Use
	an explicit check for isIntegerObject so we can return -1 on error (the stackIntegerValue:
	method answers 1 on error, and 1 is a valid pid number)."

	| result pidToSignal |
	self export: true.
	self var: 'pidToSignal' type: 'pid_t'.
	self var: 'result' type: 'int'.

	"Do not allow signal sending if running in secure mode"
	(self sandboxSecurity == 1)
		ifTrue:
			[interpreterProxy pop: 2; pushInteger: -1]
		ifFalse:
			[(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0))
				ifTrue: 
					[pidToSignal _ interpreterProxy stackIntegerValue: 0.
					result _ self cCode: 'kill(pidToSignal, SIGUSR2)' inSmalltalk: [-1].
					interpreterProxy pop: 2; pushInteger: result]
				ifFalse: [interpreterProxy pop: 2; pushInteger: -1]]! !

!UnixOSProcessPlugin methodsFor: 'primitives - signal forwarding' stamp: 'dtl 2/8/2003 09:51'!
primitiveForwardSignalToSemaphore
	"Set a signal handler in the VM which will signal a Smalltalk semaphore at
	semaphoreIndex whenever an external signal sigNum is received. Answer the
	prior value of the signal handler. If semaphoreIndex is zero or nil, the handler
	is unregistered, and the VM returns to its default behavior for handling that
	signal.

	The Smalltalk semaphore is expected to be kept at the same index location
	indefinitely during the lifetime of a Squeak session. If that is not the case, the
	handler must be unregistered prior to unregistering the Smalltalk semaphore."

	| sigNum index semaphoreIndex handler priorHandler priorHandlerObject hPtr idx |
	self export: true.
	self var: 'handler' declareC: 'void *handler'.
	self var: 'priorHandler' declareC: 'union {void *handler; unsigned char bytes[sizeof(void *)];} priorHandler'.
	self var: 'hPtr' declareC: 'char *hPtr'.
	index _ interpreterProxy stackValue: 0.
	(index == interpreterProxy nilObject)
		ifTrue:
			[semaphoreIndex _ 0]
		ifFalse:
			[(interpreterProxy isIntegerObject: index)
				ifTrue:
					[semaphoreIndex _ interpreterProxy stackIntegerValue: 0]
				ifFalse:
					[^ interpreterProxy primitiveFail]].
	sigNum _ interpreterProxy stackIntegerValue: 1.
	handler _ self forwardSignal: sigNum toSemaphoreAt: semaphoreIndex.
	(handler == self sigErrorNumber) ifTrue: [^ interpreterProxy primitiveFail].
	priorHandlerObject _ interpreterProxy
		instantiateClass: interpreterProxy classByteArray
		indexableSize: self sizeOfPointer.
	hPtr _ interpreterProxy arrayValueOf: priorHandlerObject.
	self cCode: 'priorHandler.handler = handler'.
	idx _ 0.
	[idx < self sizeOfPointer] whileTrue:
		[self cCode: 'hPtr[idx] = priorHandler.bytes[idx]'.
		idx _ idx + 1].
	interpreterProxy pop: 3 thenPush: priorHandlerObject

! !

!UnixOSProcessPlugin methodsFor: 'primitives - signal forwarding' stamp: 'dtl 8/4/2002 01:16'!
primitiveSemaIndexFor
	"Answer the registration index of the semaphore currently associated with the
	signal handler for sigNum."

	| sigNum index |
	self export: true.
	sigNum _ interpreterProxy stackIntegerValue: 0.
	index _ self semaphoreIndices at: sigNum.
	interpreterProxy pop: 1; pushInteger: index
! !

!UnixOSProcessPlugin methodsFor: 'primitives - signal forwarding' stamp: 'dtl 8/4/2002 00:26'!
primitiveSigChldNumber
	"Integer value corresponding to SIGCHLD"

	self export: true.
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: self sigChldNumber
! !

!UnixOSProcessPlugin methodsFor: 'primitives - signal forwarding' stamp: 'dtl 12/6/2002 17:56'!
primitiveSigHupNumber
	"Integer value corresponding to SIGHUP"

	self export: true.
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: self sigHupNumber
! !

!UnixOSProcessPlugin methodsFor: 'primitives - signal forwarding' stamp: 'dtl 8/4/2002 00:26'!
primitiveSigIntNumber
	"Integer value corresponding to SIGINT"

	self export: true.
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: self sigIntNumber
! !

!UnixOSProcessPlugin methodsFor: 'primitives - signal forwarding' stamp: 'dtl 12/6/2002 17:57'!
primitiveSigKillNumber
	"Integer value corresponding to SIGKILL"

	self export: true.
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: self sigKillNumber
! !

!UnixOSProcessPlugin methodsFor: 'primitives - signal forwarding' stamp: 'dtl 8/4/2002 00:27'!
primitiveSigPipeNumber
	"Integer value corresponding to SIGPIPE"

	self export: true.
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: self sigPipeNumber
! !

!UnixOSProcessPlugin methodsFor: 'primitives - signal forwarding' stamp: 'dtl 12/6/2002 17:57'!
primitiveSigQuitNumber
	"Integer value corresponding to SIGQUIT"

	self export: true.
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: self sigQuitNumber
! !

!UnixOSProcessPlugin methodsFor: 'primitives - signal forwarding' stamp: 'dtl 12/6/2002 17:58'!
primitiveSigTermNumber
	"Integer value corresponding to SIGTERM"

	self export: true.
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: self sigTermNumber
! !

!UnixOSProcessPlugin methodsFor: 'primitives - signal forwarding' stamp: 'dtl 11/4/2005 06:39'!
primitiveSigUsr1Number
	"Integer value corresponding to SIGUSR1"

	self export: true.
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: self sigUsr1Number
! !

!UnixOSProcessPlugin methodsFor: 'primitives - signal forwarding' stamp: 'dtl 11/4/2005 06:40'!
primitiveSigUsr2Number
	"Integer value corresponding to SIGUSR2"

	self export: true.
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: self sigUsr2Number
! !

!UnixOSProcessPlugin methodsFor: 'private - pipes' stamp: 'dtl 1/31/2004 18:47'!
createPipeForReader: readerIOStreamPtr writer: writerIOStreamPtr
	"Create a pipe and populate the readerIOStream and writerIOStream variables.
	The SIGPIPE handler must have been set before creating the pipe. Answer true for
	success, else false."

	| filedes |
	self var: 'readerIOStreamPtr' type: 'FILEHANDLETYPE *'.
	self var: 'writerIOStreamPtr' type: 'FILEHANDLETYPE *'.
	self var: 'filedes' declareC: 'int filedes[2]'.
	((self pipe: filedes) = -1) "Translates to a pipe() system call"
		ifTrue:
			[^ false]
		ifFalse:
			[self cCode: '*writerIOStreamPtr= (FILE *) fdopen (filedes[1], "a")'.
			self cCode: '*readerIOStreamPtr= (FILE *) fdopen (filedes[0], "r")'.
			^ true]
! !

!UnixOSProcessPlugin methodsFor: 'private - file handles' stamp: 'dtl 2/24/2001 11:39'!
descriptorTableSize
	"Answer the size of the file descriptor table for a process. I am not sure of the most portable
	way to do this. If this implementation does not work on your Unix platform, try changing
	it to answer the value of FOPEN_MAX, which will hopefully be defined in stdio.h. If
	all else fails, just hard code it to answer 20, which would be safe for any Unix."

	^ self cCode: 'getdtablesize()' inSmalltalk: [20]
! !

!UnixOSProcessPlugin methodsFor: 'private - file handles' stamp: 'dtl 1/30/2003 22:42'!
dupToStdErr: anSQFileDataStructure
	"Dup a file descriptor to allow it to be attached as the standard error when we
	exec() a new executable. This is Unix specific, in that it assumes that file descriptor
	0 is stdin, 1 is stdout, and 2 is stderr. The dup2() call is used to copy the open file
	descriptors into file descriptors 0, 1 and 2 so that the image which we execute will
	use them as stdin, stdout, and stderr."

	| filenoToDup |
	self returnTypeC: 'void'.
	filenoToDup _ self fileDescriptorFrom: anSQFileDataStructure.
	(filenoToDup < 0) ifFalse: [
		(filenoToDup = 2) ifFalse: [
			self cCode: 'fflush(stderr)'.
			self cCode: 'dup2(filenoToDup, 2)' ]]! !

!UnixOSProcessPlugin methodsFor: 'private - file handles' stamp: 'dtl 1/30/2003 22:42'!
dupToStdIn: anSQFileDataStructure
	"Dup a file descriptor to allow it to be attached as the standard input when we
	exec() a new executable. This is Unix specific, in that it assumes that file descriptor
	0 is stdin, 1 is stdout, and 2 is stderr. The dup2() call is used to copy the open file
	descriptors into file descriptors 0, 1 and 2 so that the image which we execute will
	use them as stdin, stdout, and stderr."

	| filenoToDup |
	self returnTypeC: 'void'.
	filenoToDup _ self fileDescriptorFrom: anSQFileDataStructure.
	(filenoToDup < 0) ifFalse: [
		(filenoToDup = 0) ifFalse: [
			self cCode: 'fflush(stdin)'.
			self cCode: 'dup2(filenoToDup, 0)'.
			self cCode: 'rewind(stdin)' ]]		"Set stream positioning for stdin."! !

!UnixOSProcessPlugin methodsFor: 'private - file handles' stamp: 'dtl 1/30/2003 22:42'!
dupToStdOut: anSQFileDataStructure
	"Dup a file descriptor to allow it to be attached as the standard output when we
	exec() a new executable. This is Unix specific, in that it assumes that file descriptor
	0 is stdin, 1 is stdout, and 2 is stderr. The dup2() call is used to copy the open file
	descriptors into file descriptors 0, 1 and 2 so that the image which we execute will
	use them as stdin, stdout, and stderr."

	| filenoToDup |
	self returnTypeC: 'void'.
	filenoToDup _ self fileDescriptorFrom: anSQFileDataStructure.
	(filenoToDup < 0) ifFalse: [
		(filenoToDup = 1) ifFalse: [
			self cCode: 'fflush(stdout)'.
			self cCode: 'dup2(filenoToDup, 1)' ]]
! !

!UnixOSProcessPlugin methodsFor: 'private - file handles' stamp: 'dtl 9/16/2002 23:44'!
fileDescriptorFrom: aFileHandle
	"Answer the OS file descriptor, an integer value, from a SQFile data structure,
	or answer -1 if unable to obtain the file descriptor (probably due to receiving
	an incorrect type of object as aFileHandle). This method may be called from a
	primitive, and is not intended to be called from Smalltalk."

	| sqFile osFileStream |
	self var: 'sqFile' declareC: 'SQFile *sqFile'.
	self var: 'osFileStream' declareC: 'FILE *osFileStream'.
	sqFile _ self fileValueOf: aFileHandle.
	sqFile = 0
		ifTrue:
			[^ -1]	"Something is wrong, bail out before we core dump."
		ifFalse:
			[osFileStream _ self cCode: 'sqFile->file'.
			osFileStream = 0
				ifTrue:
					[^ -1].	"Something is wrong, bail out before we core dump."
			^ self fileno: osFileStream]
! !

!UnixOSProcessPlugin methodsFor: 'private - signal handers' stamp: 'dtl 2/1/2004 12:02'!
forwardSignal: sigNum toSemaphoreAt: semaphoreIndex
	"Set a signal handler in the VM which will signal a Smalltalk semaphore at
	semaphoreIndex whenever an external signal sigNum is received. Answer the
	prior value of the signal handler. If semaphoreIndex is zero, the handler is
	unregistered, and the VM returns to its default behavior for handling that
	signal. A handler must be unregistered before it can be registered again.

	The Smalltalk semaphore is expected to be kept at the same index location
	indefinitely during the lifetime of a Squeak session. If that is not the case, the
	handler must be unregistered prior to unregistering the Smalltalk semaphore."

	| oldHandler |
	self returnTypeC: 'void *'.
	self var: 'oldHandler' type: 'void *'.
	(semaphoreIndex == 0)
		ifTrue:
			["Disable the handler"
			((self semaphoreIndices at: sigNum) > 0)
				ifTrue:
					[oldHandler _ self originalSignalHandlers at: sigNum.
					self setSignalNumber: sigNum handler: oldHandler.
					self semaphoreIndices at: sigNum put: 0.
					^ oldHandler]
				ifFalse:
					["Signal handler had not been set, answer an error"
					^ self sigErrorNumber]].
	((self semaphoreIndices at: sigNum) > 0)
		ifTrue:
			["Handler is already set, answer an error"
			^ self sigErrorNumber].
	oldHandler _ self setSignalNumber: sigNum handler: (self handleSignalFunctionAddress).
	(oldHandler ~= self sigErrorNumber)
		ifTrue:
			[self originalSignalHandlers at: sigNum put: oldHandler.
			self semaphoreIndices at: sigNum put: semaphoreIndex].
	^ oldHandler
! !

!UnixOSProcessPlugin methodsFor: 'private - signal handers' stamp: 'dtl 8/3/2002 18:54'!
handleSignalFunctionAddress

	self returnTypeC: 'void *'.
	^ self cCode: 'handleSignal' inSmalltalk: [#handleSignal:].
! !

!UnixOSProcessPlugin methodsFor: 'private - signal handers' stamp: 'dtl 6/12/2005 23:24'!
handleSignal: sigNum
	"This is a signal handler function which runs when a signal is received from the
	operating system. When the signal is received, a Smalltalk Semaphore is signaled.
	This effectively passes the external signal to Squeak to allow it to be handled in
	Smalltalk.

	Note the this handler runs the VM in the context of a Unix signal handler in order
	to signal the Smalltalk Semaphore. In a single threaded Squeak VM, this works fine,
	however it may not work as expected if pthreads are used for the VM or for any
	plugins linked internally with the VM. On GCC systems, compiling with -D_REENTRANT
	seems to produce the intended behavior."

	| semaIndex |
	self returnTypeC: 'void'.
	self var: 'sigNum' type: 'int'.
	semaIndex _ self semaphoreIndices at: sigNum.
	self forwardSignal: sigNum toSemaphoreAt: semaIndex. "Reset handler"
	(semaIndex > 0) ifTrue: [interpreterProxy signalSemaphoreWithIndex: semaIndex]
! !

!UnixOSProcessPlugin methodsFor: 'private - signal handers' stamp: 'dtl 1/31/2004 19:57'!
restoreDefaultSignalHandlers
	"Restore signal handlers to their original behaviors."

	| sigNum |
	self returnTypeC: 'void'.
	sigNum _ 1.
	[sigNum <= self signalArraySize] whileTrue:
		[((self semaphoreIndices at: sigNum) > 0) ifTrue:
			[self setSignalNumber: sigNum handler: (self originalSignalHandlers at: sigNum).].
		sigNum _ sigNum + 1]
! !

!UnixOSProcessPlugin methodsFor: 'private - signal handers' stamp: 'dtl 1/25/2004 19:23'!
setSignalNumber: anInteger handler: signalHandlerAddress
	"Set a signal handler. The C code translator will convert #sig:nal: into 'signal(parm1, parm2)'"

	| oldHandlerAddress |
	self returnTypeC: 'void *'.
	self var: 'signalHandlerAddress' type: 'void *'.
	^ self sig: anInteger nal: signalHandlerAddress
! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl 8/3/2002 14:46'!
sigChldNumber
	"Child status has changed (POSIX)."

	^ self cCode: 'SIGCHLD' inSmalltalk: [17]
! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl 2/8/2003 09:50'!
sigDefaultNumber
	"Default action for a signal"

	self returnTypeC: 'void *'.
	^ self cCode: 'SIG_DFL' inSmalltalk: [0]
! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl 2/8/2003 09:50'!
sigErrorNumber
	"Error return from signal()"

	self returnTypeC: 'void *'.
	^ self cCode: 'SIG_ERR' inSmalltalk: [-1]
! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl 12/6/2002 17:47'!
sigHupNumber
	"Hangup detected on controlling terminal or death of controlling process"

	^ self cCode: 'SIGHUP' inSmalltalk: [1]
! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl 7/12/2003 11:07'!
sigIgnoreNumber
	"Ignore action for a signal"

	self returnTypeC: 'void *'.
	^ self cCode: 'SIG_IGN' inSmalltalk: [1]
! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl 8/3/2002 14:46'!
sigIntNumber
	"Interrupt (ANSI)."

	^ self cCode: 'SIGINT' inSmalltalk: [2]
! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl 12/6/2002 17:49'!
sigKillNumber
	"Kill signal"

	^ self cCode: 'SIGKILL' inSmalltalk: [9]
! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl 8/3/2002 14:46'!
sigPipeNumber
	"Broken pipe (POSIX)."

	^ self cCode: 'SIGPIPE' inSmalltalk: [13]
! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl 12/6/2002 17:48'!
sigQuitNumber
	"Quit from keyboard"

	^ self cCode: 'SIGQUIT' inSmalltalk: [3]
! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl 12/6/2002 17:46'!
sigTermNumber
	"Termination signal. This is the default signal sent by the unix kill(1) command."

	^ self cCode: 'SIGTERM' inSmalltalk: [15]
! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl 11/4/2005 06:39'!
sigUsr1Number
	"User defined signal number 1. This is value is platform-dependent, so the
	inSmalltalk default of 10 may be wrong on some platforms."

	^ self cCode: 'SIGUSR1' inSmalltalk: [10]
! !

!UnixOSProcessPlugin methodsFor: 'private - signum' stamp: 'dtl 11/4/2005 06:39'!
sigUsr2Number
	"User defined signal number 2. This is value is platform-dependent, so the
	inSmalltalk default of 12 may be wrong on some platforms."

	^ self cCode: 'SIGUSR2' inSmalltalk: [12]
! !

!UnixOSProcessPlugin methodsFor: 'deprecated' stamp: 'dtl 8/7/2005 13:59'!
aioForward: fd withData: data andFlags: flags

	"Deprecated. This primitive has been replaced by a similar implementation in AioPlugin."

	"This function is called to signal a Smalltalk Semaphore when an asynchronous event is
	detected. When translated to C, the name of this method is aioForwardwithDataandFlags.
	The event handler is set up by #primitiveAioHandle."

	| semaIndex pfd |
	self returnTypeC: 'void'.
	self var: 'fd' type: 'int'.
	self var: 'data' type: 'void *'.
	self var: 'flags' type: 'int'.
	self var: 'pfd' declareC: 'sqInt *pfd'.
	pfd _ data.
	semaIndex _ self cCode: '*pfd' inSmalltalk: [0].
	interpreterProxy signalSemaphoreWithIndex: semaIndex
! !

!UnixOSProcessPlugin methodsFor: 'deprecated' stamp: 'dtl 8/4/2002 12:37'!
makePipeForReader: readerIOStreamPtr writer: writerIOStreamPtr
	"Create a pipe and populate the readerIOStream and writerIOStream variables.
	Make sure the SIGPIPE handler is set before creating the pipe. Answer true for
	success, else false"

	| filedes |
	self var: 'readerIOStreamPtr' type: 'FILEHANDLETYPE *'.
	self var: 'writerIOStreamPtr' type: 'FILEHANDLETYPE *'.
	self var: 'filedes' declareC: 'int filedes[2]'.
	self setSigPipeHandler.
	((self pipe: filedes) = -1) "Translates to a pipe() system call"
		ifTrue:
			[^ false]
		ifFalse:
			[self cCode: '*writerIOStreamPtr= (FILE *) fdopen (filedes[1], "a")'.
			self cCode: '*readerIOStreamPtr= (FILE *) fdopen (filedes[0], "r")'.
			^ true]
! !

!UnixOSProcessPlugin methodsFor: 'deprecated' stamp: 'dtl 8/7/2005 13:59'!
primitiveAioDisable

	"Deprecated. This primitive has been replaced by a similar implementation in AioPlugin."

	"Definitively disable asynchronous event notification for a descriptor. The
	parameter is a sqFile ByteArray representing the ioHandle for a file."

	| sqFile fd |
	self export: true.
	self var: 'fd' type: 'int'.
	sqFile _ interpreterProxy stackObjectValue: 0.
	fd _ self fileDescriptorFrom: sqFile.
	(fd < 0) ifTrue: [^ interpreterProxy primitiveFail].
	self aioDisable: fd.
	interpreterProxy pop: 2; pushInteger: fd
! !

!UnixOSProcessPlugin methodsFor: 'deprecated' stamp: 'dtl 8/7/2005 13:59'!
primitiveAioEnable

	"Deprecated. This primitive has been replaced by a similar implementation in AioPlugin."

	"Enable asynchronous notification for a descriptor. The first parameter is a sqFile
	ByteArray representing the ioHandle for a file, the second parameter is the index
	of a Semaphore to be notified, and the third parameter is a flag indicating that
	sqFile represents an external object and should not be closed on termination of
	aio handling. Answer the semaphore index."

	| sqFile fd semaIndex flags externalObject eventSemaphoreIndices |
	self export: true.
	self var: 'fd' type: 'int'.
	self var: 'flags' type: 'int'.
	self var: 'eventSemaphoreIndices' declareC: 'static int eventSemaphoreIndices[FD_SETSIZE]'.
	sqFile _ interpreterProxy stackObjectValue: 2.
	((interpreterProxy isBytes: sqFile) and: [(interpreterProxy stSizeOf: sqFile) = self fileRecordSize])
		ifFalse: [^ interpreterProxy primitiveFail].
	fd _ self fileDescriptorFrom: sqFile.
	(fd < 0) ifTrue: [^ interpreterProxy primitiveFail].
	semaIndex _ interpreterProxy stackIntegerValue: 1.
	"The following is to prevent the need to coerce integers into pointers in clientData".
	eventSemaphoreIndices at: semaIndex put: semaIndex.
	externalObject _ interpreterProxy stackObjectValue: 0.
	(externalObject = interpreterProxy trueObject)
		ifTrue: [flags _ self cCode: 'AIO_EXT' inSmalltalk: [16]]
		ifFalse: [flags _ 0].
	self cCode: 'aioEnable(fd, &(eventSemaphoreIndices[semaIndex]), flags)'.
	interpreterProxy pop: 4; pushInteger: semaIndex.! !

!UnixOSProcessPlugin methodsFor: 'deprecated' stamp: 'dtl 8/7/2005 13:59'!
primitiveAioHandle

	"Deprecated. This primitive has been replaced by a similar implementation in AioPlugin."

	"Handle asynchronous event notification for a descriptor. The first parameter is
	a sqFile ByteArray representing the ioHandle for a file, and the remaining three
	parameters are Boolean flags representing the types of events for which notification
	is being requested: handle exceptions, handle for read, and handle for write.
	Flags are defined in the aio.h source as:
		AIO_X	(1<<0)	handle for exceptions
		AIO_R	(1<<1)	handle for read
		AIO_W	(1<<2)	handle for write"

	| sqFile fd exceptionWatch readWatch writeWatch flags |
	self export: true.
	self var: 'fd' type: 'int'.
	self var: 'flags' type: 'int'.
	sqFile _ interpreterProxy stackObjectValue: 3.
	((interpreterProxy isBytes: sqFile) and: [(interpreterProxy stSizeOf: sqFile) = self fileRecordSize])
		ifFalse: [^ interpreterProxy primitiveFail].
	fd _ self fileDescriptorFrom: sqFile.
	(fd < 0) ifTrue: [^ interpreterProxy primitiveFail].
	exceptionWatch _ interpreterProxy stackObjectValue: 2.
	readWatch _ interpreterProxy stackObjectValue: 1.
	writeWatch _ interpreterProxy stackObjectValue: 0.
	flags _ 0.
	(exceptionWatch == interpreterProxy trueObject)
		ifTrue: [flags _ flags bitOr: (self cCode: 'AIO_X' inSmalltalk: [1])].
	(readWatch == interpreterProxy trueObject)
		ifTrue: [flags _ flags bitOr: (self cCode: 'AIO_R' inSmalltalk: [2])].
	(writeWatch == interpreterProxy trueObject)
		ifTrue: [flags _ flags bitOr: (self cCode: 'AIO_W' inSmalltalk: [4])].
	self cCode: 'aioHandle(fd, aioForwardwithDataandFlags, flags)'.
	interpreterProxy pop: 5; pushInteger: flags
! !

!UnixOSProcessPlugin methodsFor: 'deprecated' stamp: 'dtl 8/7/2005 13:59'!
primitiveAioSuspend

	"Deprecated. This primitive has been replaced by a similar implementation in AioPlugin."

	"Temporarily suspend asynchronous event notification for a descriptor. The first
	parameter is a sqFile ByteArray representing the ioHandle for a file, and the
	remaining three parameters are Boolean flags representing the types of events
	for which notification is being requested: handle exceptions, handle for read,
	and handle for write.
	Flags are defined in the aio.h source as:
		AIO_X	(1<<0)	handle for exceptions
		AIO_R	(1<<1)	handle for read
		AIO_W	(1<<2)	handle for write"

	| sqFile fd exceptionWatch readWatch writeWatch flags |
	self export: true.
	self var: 'fd' type: 'int'.
	self var: 'flags' type: 'int'.
	sqFile _ interpreterProxy stackObjectValue: 3.
	((interpreterProxy isBytes: sqFile) and: [(interpreterProxy stSizeOf: sqFile) = self fileRecordSize])
		ifFalse: [^ interpreterProxy primitiveFail].
	fd _ self fileDescriptorFrom: sqFile.
	(fd < 0) ifTrue: [^ interpreterProxy primitiveFail].
	exceptionWatch _ interpreterProxy stackObjectValue: 2.
	readWatch _ interpreterProxy stackObjectValue: 1.
	writeWatch _ interpreterProxy stackObjectValue: 0.
	flags _ 0.
	(exceptionWatch == interpreterProxy trueObject)
		ifTrue: [flags _ flags bitOr: (self cCode: 'AIO_X' inSmalltalk: [1])].
	(readWatch == interpreterProxy trueObject)
		ifTrue: [flags _ flags bitOr: (self cCode: 'AIO_R' inSmalltalk: [2])].
	(writeWatch == interpreterProxy trueObject)
		ifTrue: [flags _ flags bitOr: (self cCode: 'AIO_W' inSmalltalk: [4])].
	self cCode: 'aioSuspend(fd, flags)'.
	interpreterProxy pop: 5; pushInteger: flags
! !

!UnixOSProcessPlugin methodsFor: 'deprecated' stamp: 'dtl 1/25/2004 12:29'!
primitiveForkAndExecInDirectory
	"Fork a child OS process, and do an exec in the child. The parent continues on in
	Smalltalk, and this method answers the pid of the child which was created.

	On entry, the stack contains:
		0: workingDir, a null terminated string specifying the working directory to use, or nil.
		1: envOffsets, an array of integers for calculating environment string address offsets.
		2: envVecBuffer, a String buffer containing environment strings arranged to look like char **.
		3: argOffsets, an array of integers for calculating argument string address offsets.
		4: argVecBuffer, a String buffer containing argument strings arranged to look like char **.
		5: stdErr, a ByteArray handle.
		6: stdOut, a ByteArray handle.
		7: stdIn, a ByteArray handle.
		8: executableFile, a null terminated string with the name of the file to execute.
		9: the sender"

	self export: true.
	^ self forkAndExecInDirectory: true
! !

!UnixOSProcessPlugin methodsFor: 'deprecated' stamp: 'dtl 2/21/2002 15:44'!
primitiveSetSemaIndex
	"Set the index of the semaphore used by the OSProcess with which I collaborate. My
	OSProcess should set this value so that I can use it when handling SIGCHLD signals
	(death of child). In the C translation this is a static int which would be shared by all
	instances of UnixOSProcessPlugin, which is expected to be a singleton. Answer the
	value of the semaphore index."

	self export: true.
	sigChldSemaIndex _ interpreterProxy stackIntegerValue: 0.
	interpreterProxy pop: 2; pushInteger: sigChldSemaIndex
! !

!UnixOSProcessPlugin methodsFor: 'deprecated' stamp: 'dtl 6/12/2005 15:33'!
reapChildProcess: sigNum
	"This is a signal handler for SIGCHLD. It is not meant to be called from Smalltalk,
	and should only be called indirectly as a result of a death of child signal from
	the operating system.

	Child processes must be cleaned up by the parent, otherwise they continue
	to exist as zombies until the parent exits. This handler resets the signal handler
	to catch the next SIGCHLD signal, then sets a semaphore to notify the system
	that a child process needs to be cleaned up. The actual clean up is done by a
	Smalltalk process which waits on the semaphore, then calls primitiveReapChildProcess.

	Note: If child processes die faster than we can clean them up, signals will be lost
	and child processes will remain as zombies."

	self returnTypeC: 'void'.
	self var: 'sigNum' type: 'int'.
	self setSigChldHandler.	"Reset handler"
	(sigChldSemaIndex > 0)
		ifTrue: [interpreterProxy signalSemaphoreWithIndex: sigChldSemaIndex]
! !

!UnixOSProcessPlugin methodsFor: 'deprecated' stamp: 'dtl 2/8/2003 09:50'!
setSigChldDefaultHandler

	self returnTypeC: 'void'.
	self setSignalNumber: self sigChldNumber handler: self sigDefaultNumber
! !

!UnixOSProcessPlugin methodsFor: 'deprecated' stamp: 'dtl 2/8/2003 09:51'!
setSigChldHandler
	"Set the SIGCHLD signal handler in the virtual machine."

	self returnTypeC: 'void'.
	((self cCode: 'signal(SIGCHLD, reapChildProcess)' inSmalltalk: [self sigErrorNumber]) == self sigErrorNumber)
		ifTrue: [self perror: 'signal']
! !

!UnixOSProcessPlugin methodsFor: 'deprecated' stamp: 'dtl 2/8/2003 09:50'!
setSigIntDefaultHandler

	self returnTypeC: 'void'.
	self setSignalNumber: self sigIntNumber handler: self sigDefaultNumber
! !

!UnixOSProcessPlugin methodsFor: 'deprecated' stamp: 'dtl 2/8/2003 09:52'!
setSigIntIgnore
	"Set the SIGINT signal handler in the virtual machine to ignore interrupts."

	self returnTypeC: 'void'.
	self setSignalNumber: self sigIntNumber handler: self sigIgnoreNumber
! !

!UnixOSProcessPlugin methodsFor: 'deprecated' stamp: 'dtl 2/8/2003 09:50'!
setSigPipeDefaultHandler

	self returnTypeC: 'void'.
	self setSignalNumber: self sigPipeNumber handler: self sigDefaultNumber
! !

!UnixOSProcessPlugin methodsFor: 'deprecated' stamp: 'dtl 8/3/2002 14:11'!
setSigPipeHandler
	"Set the SIGPIPE signal handler in the virtual machine to ignore pipe error signals.
	If a pipe is opened to a child process, and the child exits, then subsequent writes to
	the pipe generate a SIGPIPE signal. If this signal is not handled, the VM will exit
	without warning."

	self setSigPipeIgnore
! !

!UnixOSProcessPlugin methodsFor: 'deprecated' stamp: 'dtl 2/8/2003 09:52'!
setSigPipeIgnore
	"Set the SIGPIPE signal handler in the virtual machine to ignore pipe error signals."

	self returnTypeC: 'void'.
	self setSignalNumber: self sigPipeNumber handler: self sigIgnoreNumber
! !


!UnixOSProcessPlugin class methodsFor: 'translation' stamp: 'dtl 9/17/2005 10:54'!
declareCVarsIn: cg

	"Additional declarations are done in my subclasses. This permits a meaningful
	module identification string to be written with the name of the actual subclass
	being generated."

	cg addHeaderFile: '<sys/time.h>'.
	cg addHeaderFile: '<sys/wait.h>'.
	cg addHeaderFile: '<signal.h>'.
	cg addHeaderFile: '<unistd.h>'.
	cg addHeaderFile: '<fcntl.h>'.
	cg addHeaderFile: '<sys/stat.h>'.	"For stat(2) call in primitiveFileProtectionMask"
	cg addHeaderFile: '<errno.h>'.
	cg addHeaderFile: '"config.h"
#define FILEHANDLETYPE FILE *  /* the type of low level stream to be used in a struct SQFile */
#ifndef SQAIO_H
# define SQAIO_H "aio.h"          /* aio.h has been renamed to sqaio.h */
#endif'.
	cg addHeaderFile: 'SQAIO_H'.
	cg var: 'sigHandlers' declareC: 'void *sigHandlers[NSIG]'.
	cg var: 'originalSigHandlers' declareC: 'void *originalSigHandlers[NSIG]'.
	cg var: 'semaIndices' declareC: 'unsigned char semaIndices[NSIG]'

! !

!UnixOSProcessPlugin class methodsFor: 'translation' stamp: 'dtl 3/5/2005 12:10'!
isResponsibleForThisPlatform
	"Answer true is this is an instance of the class which is responsible for representing
	the OS process for the Squeak VM running on the current platform. A false answer is
	usually the result of running the image on a different platform and VM.
	Note: Keep this method is sync with OSProcess>>isUnix."

	| numericOsVersion |

	^ (self platformName = 'unix') or:
		[numericOsVersion _ Smalltalk osVersion asInteger ifNil: [0].
		(self platformName = 'Mac OS') and: [numericOsVersion >= 1000]]
! !

!UnixOSProcessPlugin class methodsFor: 'translation' stamp: 'dtl 7/22/2000 17:15'!
moduleName

	^ 'UnixOSProcessPlugin'! !

!UnixOSProcessPlugin class methodsFor: 'translation' stamp: 'dtl 3/3/2002 11:26'!
shouldBeTranslated

	^ self isResponsibleForThisPlatform and: [self concreteClass == self]
! !

!UnixOSProcessPlugin class methodsFor: 'concrete subclasses' stamp: 'dtl 8/7/2005 22:45'!
concreteClass
	"UnixOSProcessPlugin concreteClass"

	^ self hasInterpreterGetThisSession
		ifTrue: [UnixOSProcessPluginInterpreterGetThisSession]
		ifFalse: [UnixOSProcessPluginNoThisSessionAvailable]
! !

!UnixOSProcessPlugin class methodsFor: 'version testing' stamp: 'dtl 3/1/2002 17:05'!
installedModuleName
	"Answer the module name string for the plugin which is actually running in the VM"

	"UnixOSProcessPlugin installedModuleName"

	<primitive: 'primitiveModuleName' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessPlugin class methodsFor: 'version testing' stamp: 'dtl 3/1/2002 17:24'!
installedModuleVersionString
	"Answer the version string for the plugin which is actually running in the VM"

	"UnixOSProcessPlugin installedModuleVersionString"

	<primitive: 'primitiveVersionString' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessPlugin class methodsFor: 'version testing' stamp: 'dtl 8/13/2003 21:54'!
primGetPid
	"Answer the OS process ID for the OS process in which I am currently executing."

	<primitive: 'primitiveGetPid' module: 'UnixOSProcessPlugin'>
	^ nil! !


!UnixOSProcessPluginInterpreterGetThisSession methodsFor: 'private' stamp: 'dtl 8/8/2005 06:25'!
getThisSessionIdentifier

	^ interpreterProxy getThisSessionID
! !


!UnixOSProcessPluginInterpreterGetThisSession class methodsFor: 'translation' stamp: 'dtl 8/8/2005 06:20'!
declareCVarsIn: cg

	cg addHeaderFile: '<sys/types.h>
/* ', self idString, ' - ', self moduleName, '.c translated from class
   ', self name, ' of OSProcessPlugin version ', self versionString, ' */'
! !


!UnixOSProcessPluginNoThisSessionAvailable methodsFor: 'private' stamp: 'dtl 3/3/2002 09:10'!
getThisSessionIdentifier

	^ nil! !


!UnixOSProcessPluginNoThisSessionAvailable class methodsFor: 'translation' stamp: 'dtl 8/8/2005 06:20'!
declareCVarsIn: cg

	cg addHeaderFile: '<sys/types.h>
/* ', self idString, ' - ', self moduleName, '.c translated from class
   ', self name, ' of OSProcessPlugin version ', self versionString, ' */'
! !


!Win32OSProcessPlugin methodsFor: 'initialize-release' stamp: 'dtl 8/6/2003 05:16'!
shutdownModule

	self export: true.
! !

!Win32OSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 9/6/2002 14:50'!
primitiveGetCurrentWorkingDirectory
	"Answer a string containing the current working directory."

	| buffer charSize cwdString requiredLen cwdLen |
	self export: true.
	self var: 'buffer' declareC: 'char *buffer'.
	self var: 'requiredLen' declareC: 'DWORD requiredLen'.
	self var: 'cwdLen' declareC: 'DWORD cwdLen'.

	charSize _ self cCode: 'sizeof(TCHAR)' inSmalltalk: [1].

	"Allocate a dummy buffer buffer with no space (force a failure to GetCurrentDirectory)"
	cwdString _ interpreterProxy
		instantiateClass: interpreterProxy classString
		indexableSize: 0.
	buffer _ interpreterProxy arrayValueOf: cwdString.

	"Determine required size of result buffer"
	requiredLen _ self cCode: 'GetCurrentDirectory(0, buffer)' inSmalltalk: [0].
	(requiredLen == 0) ifTrue: [^ interpreterProxy primitiveFail].

	"Allocate the real buffer"
	cwdString _ interpreterProxy
		instantiateClass: interpreterProxy classString
		indexableSize: (requiredLen * charSize).
	buffer _ interpreterProxy arrayValueOf: cwdString.

	"Call GetCurrentDirectory() again to get the actual result"
	cwdLen _ self cCode: 'GetCurrentDirectory(requiredLen, buffer)' inSmalltalk: [0].
	(cwdLen == 0) ifTrue: [^ interpreterProxy primitiveFail].
	(cwdLen > requiredLen) ifTrue: [^ interpreterProxy primitiveFail].

	interpreterProxy pop: 1; push: cwdString
 ! !

!Win32OSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 3/17/2002 09:15'!
primitiveGetEnvironmentStrings
	"Answer the environment block in the form of an Array of Strings. The
	caller is expected to parse the strings into a dictionary of keys and values."

	| pVarBlock p envArray count string |
	self export: true.
	self var: 'pVarBlock' declareC: 'TCHAR *pVarBlock'.
	self var: 'p' declareC: 'TCHAR *p'.
	pVarBlock _ self GetEnvironmentStrings.

	"Count the environment strings so the result array can be pre-allocated.
	This minimizes the number of objects which must be pushed onto the
	stack of remappable objects."
	p _ pVarBlock.
	count _ 0.
	[self cCode: '*p !!= 0' inSmalltalk: [pVarBlock ~= nil]] whileTrue:
		[count _ count + 1.
		[self cCode: '*p !!= 0' inSmalltalk: [p ~= nil]] whileTrue: [p _ p + 1].
		p _ p + 1].
	"Allocate an array for the strings."
	envArray _ interpreterProxy
		instantiateClass: interpreterProxy classArray
		indexableSize: count.
	"Populate the array."
	p _ pVarBlock.
	count _ 0.
	[self cCode: '*p !!= 0' inSmalltalk: [pVarBlock ~= nil]] whileTrue:
		[count _ count + 1.
		interpreterProxy pushRemappableOop: envArray.
		string _ self stringFromCString: p.	"May trigger garbage collection"
		envArray _ interpreterProxy popRemappableOop.
		interpreterProxy stObject: envArray at: count put: string.
		[self cCode: '*p !!= 0' inSmalltalk: [p ~= nil]] whileTrue: [p _ p + 1].
		p _ p + 1].
	self FreeEnvironmentStrings: pVarBlock.
	interpreterProxy pop: 1 thenPush: envArray! !

!Win32OSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 2/27/2002 12:40'!
primitiveGetMainThreadHandle
	"Answer the ID of my main thread."

	| handle handleOop |
	self export: true.
	self var: 'handle' type: 'HANDLE'.
	handle _ self GetCurrentThread.
	handleOop _ self handleObjectFrom: handle.
	interpreterProxy pop: 1 thenPush: handleOop
! !

!Win32OSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 2/27/2002 12:33'!
primitiveGetMainThreadID
	"Answer the ID of my main thread."

	| tid tidOop |
	self export: true.
	self var: 'tid' declareC: 'DWORD tid'.
	tid _ self GetCurrentThreadId.
	tidOop _ interpreterProxy positive32BitIntegerFor: tid.
	interpreterProxy pop: 1 thenPush: tidOop
! !

!Win32OSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 2/22/2002 16:18'!
primitiveGetPid
	"Answer the process ID of my OS process"

	| pid pidOop |
	self export: true.
	self var: 'pid' declareC: 'DWORD pid'.
	pid _ self GetCurrentProcessId.
	pidOop _ interpreterProxy positive32BitIntegerFor: pid.
	interpreterProxy pop: 1 thenPush: pidOop
! !

!Win32OSProcessPlugin methodsFor: 'primitives - OS process access' stamp: 'dtl 2/27/2002 12:40'!
primitiveGetPidHandle
	"Answer the handle for my OS process"

	| handle handleOop |
	self export: true.
	self var: 'handle' type: 'HANDLE'.
	handle _ self GetCurrentProcess.
	handleOop _ self handleObjectFrom: handle.
	interpreterProxy pop: 1 thenPush: handleOop
! !

!Win32OSProcessPlugin methodsFor: 'primitives - process creation' stamp: 'dtl 9/18/2002 09:38'!
primitiveCommandWithInputOutputError
	"Create a process to run commandString, using the specified input, output,
	and error streams. If any of the input, output, or error streams is nil, then
	the current values of these handles in this OS process will be passed to the
	new process. Answer an Array formed from the PROCESS_INFORMATION
	structure for the new process. The caller is expected to close the process and
	thread handles when they are no longer needed. The result array contains
	hProcess, hThread, dwProcessId, dwThreadId."

	| cStringPtr commandString startUp procInfo processInformation stdErr stdOut stdIn
	inputHandle outputHandle errorHandle |
	self export: true.
	self var: 'cStringPtr' declareC: 'LPTSTR cStringPtr'.
	self var: 'startUp' declareC: 'STARTUPINFO startUp'.
	self var: 'procInfo' declareC: 'PROCESS_INFORMATION procInfo'.
	self var: 'inputHandle' declareC: 'HANDLE inputHandle'.
	self var: 'outputHandle' declareC: 'HANDLE outputHandle'.
	self var: 'errorHandle' declareC: 'HANDLE errorHandle'.
	stdErr _ interpreterProxy stackObjectValue: 0.
	stdOut _ interpreterProxy stackObjectValue: 1.
	stdIn _ interpreterProxy stackObjectValue: 2.
	commandString _ interpreterProxy stackObjectValue: 3.
	cStringPtr _ self cStringFromString: commandString.
	(stdIn = interpreterProxy nilObject)
		ifTrue: [inputHandle _ self stdInHandle]
		ifFalse: [inputHandle _ self handleFromSQFile: stdIn].
	(stdOut = interpreterProxy nilObject)
		ifTrue: [outputHandle _ self stdOutHandle]
		ifFalse: [outputHandle _ self handleFromSQFile: stdOut].
	(stdErr = interpreterProxy nilObject)
		ifTrue: [errorHandle _ self stdErrHandle]
		ifFalse: [errorHandle _ self handleFromSQFile: stdErr].
	self cCode: 'GetStartupInfo (&startUp)'.
	self cCode: 'startUp.hStdInput = inputHandle'.
	self cCode: 'startUp.hStdOutput = outputHandle'.
	self cCode: 'startUp.hStdError = errorHandle'.

	(self cCode:  'CreateProcess(NULL, cStringPtr, NULL, NULL, TRUE, NORMAL_PRIORITY_CLASS, NULL, NULL, &startUp, &procInfo)')
		ifFalse: [^ interpreterProxy primitiveFail].

	interpreterProxy pushRemappableOop: (interpreterProxy positive32BitIntegerFor: (self cCode: 'procInfo.dwThreadId')).
	interpreterProxy pushRemappableOop: (interpreterProxy positive32BitIntegerFor: (self cCode: 'procInfo.dwProcessId')).
	interpreterProxy pushRemappableOop: (self handleObjectFrom: (self cCode: 'procInfo.hThread')).
	interpreterProxy pushRemappableOop: (self handleObjectFrom: (self cCode: 'procInfo.hProcess')).
	processInformation _ interpreterProxy
		instantiateClass: interpreterProxy classArray
		indexableSize: 4.
	interpreterProxy
		storePointer: 0
		ofObject: processInformation
		withValue: interpreterProxy popRemappableOop.
	interpreterProxy
		storePointer: 1
		ofObject: processInformation
		withValue: interpreterProxy popRemappableOop.
	interpreterProxy
		storePointer: 2
		ofObject: processInformation
		withValue: interpreterProxy popRemappableOop.
	interpreterProxy
		storePointer: 3
		ofObject: processInformation
		withValue: interpreterProxy popRemappableOop.
	interpreterProxy pop: 5; push: processInformation
! !

!Win32OSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl 9/7/2002 11:13'!
primitiveGetStdError
	"Answer a two element array containing the sqFile data structure representing
	standard error stream for my OS process, and a flag (true or false) to indicate
	whether the sqFile data structure contains a valid HANDLE. If no standard error
	stream is available for this OS process, the sqFile data structure will contain an
	invalid HANDLE value, which will result in failures on subsequent accesses."

	| file handle result invalid thisSession fileOop |
	self export: true.
	self var: 'file' declareC: 'SQFile *file'.
	self var: 'handle' declareC: 'HANDLE handle'.
	self var: 'invalid' declareC: 'HANDLE invalid'.
	thisSession _ self sessionIdentifierFrom: (interpreterProxy stackObjectValue: 0).
	invalid _ self cCode: 'INVALID_HANDLE_VALUE' inSmalltalk: [-1].
	result _ interpreterProxy
		instantiateClass: interpreterProxy classArray
		indexableSize: 2.
	fileOop _ self newSQFileByteArray.
	file _ self fileValueOf: fileOop.
	handle _ self cCode: 'GetStdHandle(STD_ERROR_HANDLE)' inSmalltalk: [-1].
	(handle ~= invalid)
		ifTrue: [interpreterProxy stObject: result at: 2 put: interpreterProxy trueObject]
		ifFalse: [interpreterProxy stObject: result at: 2 put: interpreterProxy falseObject].
	self cCode: 'file->file = handle'.
	self cCode: 'file->sessionID = thisSession'.
	self cCode: 'file->writable = 1'.
	self cCode: 'file->lastOp = 0'.
	interpreterProxy stObject: result at: 1 put: fileOop.
	interpreterProxy pop: 2; push: result
! !

!Win32OSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl 9/7/2002 11:14'!
primitiveGetStdInput
	"Answer a two element array containing the sqFile data structure representing
	standard input stream for my OS process, and a flag (true or false) to indicate
	whether the sqFile data structure contains a valid HANDLE. If no standard input
	stream is available for this OS process, the sqFile data structure will contain an
	invalid HANDLE value, which will result in failures on subsequent accesses."

	| file handle result invalid thisSession fileOop |
	self export: true.
	self var: 'file' declareC: 'SQFile *file'.
	self var: 'handle' declareC: 'HANDLE handle'.
	self var: 'invalid' declareC: 'HANDLE invalid'.
	thisSession _ self sessionIdentifierFrom: (interpreterProxy stackObjectValue: 0).
	invalid _ self cCode: 'INVALID_HANDLE_VALUE' inSmalltalk: [-1].
	result _ interpreterProxy
		instantiateClass: interpreterProxy classArray
		indexableSize: 2.
	fileOop _ self newSQFileByteArray.
	file _ self fileValueOf: fileOop.
	handle _ self cCode: 'GetStdHandle(STD_INPUT_HANDLE)' inSmalltalk: [-1].
	(handle ~= invalid)
		ifTrue: [interpreterProxy stObject: result at: 2 put: interpreterProxy trueObject]
		ifFalse: [interpreterProxy stObject: result at: 2 put: interpreterProxy falseObject].
	self cCode: 'file->file = handle'.
	self cCode: 'file->sessionID = thisSession'.
	self cCode: 'file->writable = 0'.
	self cCode: 'file->lastOp = 0'.
	interpreterProxy stObject: result at: 1 put: fileOop.
	interpreterProxy pop: 2; push: result
! !

!Win32OSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl 9/7/2002 11:14'!
primitiveGetStdOutput
	"Answer a two element array containing the sqFile data structure representing
	standard output stream for my OS process, and a flag (true or false) to indicate
	whether the sqFile data structure contains a valid HANDLE. If no standard output
	stream is available for this OS process, the sqFile data structure will contain an
	invalid HANDLE value, which will result in failures on subsequent accesses."

	| file handle result invalid thisSession fileOop |
	self export: true.
	self var: 'file' declareC: 'SQFile *file'.
	self var: 'handle' declareC: 'HANDLE handle'.
	self var: 'invalid' declareC: 'HANDLE invalid'.
	thisSession _ self sessionIdentifierFrom: (interpreterProxy stackObjectValue: 0).
	invalid _ self cCode: 'INVALID_HANDLE_VALUE' inSmalltalk: [-1].
	result _ interpreterProxy
		instantiateClass: interpreterProxy classArray
		indexableSize: 2.
	fileOop _ self newSQFileByteArray.
	file _ self fileValueOf: fileOop.
	handle _ self cCode: 'GetStdHandle(STD_OUTPUT_HANDLE)' inSmalltalk: [-1].
	(handle ~= invalid)
		ifTrue: [interpreterProxy stObject: result at: 2 put: interpreterProxy trueObject]
		ifFalse: [interpreterProxy stObject: result at: 2 put: interpreterProxy falseObject].
	self cCode: 'file->file = handle'.
	self cCode: 'file->sessionID = thisSession'.
	self cCode: 'file->writable = 1'.
	self cCode: 'file->lastOp = 0'.
	interpreterProxy stObject: result at: 1 put: fileOop.
	interpreterProxy pop: 2; push: result
! !

!Win32OSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl 9/18/2002 10:35'!
primitiveSetStdErr
	"Set the file handle for standard error of my OS process to the handle passed in a SQFile struct."

	| sqFileOop sqFile handle |
	self export: true.
	self var: 'sqFile' declareC: 'SQFile *sqFile'.
	self var: 'handle' type: 'HANDLE'.
	sqFileOop _ interpreterProxy stackValue: 0.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	sqFile _ interpreterProxy arrayValueOf: sqFileOop.
	handle _ self cCode: 'sqFile->file' inSmalltalk: [0].
	(self cCode: 'SetStdHandle(STD_ERROR_HANDLE, handle)' inSmalltalk: [false])
		ifTrue: [interpreterProxy pop: 2 thenPush: interpreterProxy trueObject]
		ifFalse: [interpreterProxy pop: 2 thenPush: interpreterProxy falseObject]
! !

!Win32OSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl 9/18/2002 10:35'!
primitiveSetStdIn
	"Set the file handle for standard input of my OS process to the handle passed in a SQFile struct."

	| sqFileOop sqFile handle |
	self export: true.
	self var: 'sqFile' declareC: 'SQFile *sqFile'.
	self var: 'handle' type: 'HANDLE'.
	sqFileOop _ interpreterProxy stackValue: 0.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	sqFile _ interpreterProxy arrayValueOf: sqFileOop.
	handle _ self cCode: 'sqFile->file' inSmalltalk: [0].
	(self cCode: 'SetStdHandle(STD_INPUT_HANDLE, handle)' inSmalltalk: [false])
		ifTrue: [interpreterProxy pop: 2 thenPush: interpreterProxy trueObject]
		ifFalse: [interpreterProxy pop: 2 thenPush: interpreterProxy falseObject]
! !

!Win32OSProcessPlugin methodsFor: 'primitives - files' stamp: 'dtl 9/18/2002 10:35'!
primitiveSetStdOut
	"Set the file handle for standard output of my OS process to the handle passed in a SQFile struct."

	| sqFileOop sqFile handle |
	self export: true.
	self var: 'sqFile' declareC: 'SQFile *sqFile'.
	self var: 'handle' type: 'HANDLE'.
	sqFileOop _ interpreterProxy stackValue: 0.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	sqFile _ interpreterProxy arrayValueOf: sqFileOop.
	handle _ self cCode: 'sqFile->file' inSmalltalk: [0].
	(self cCode: 'SetStdHandle(STD_OUTPUT_HANDLE, handle)' inSmalltalk: [false])
		ifTrue: [interpreterProxy pop: 2 thenPush: interpreterProxy trueObject]
		ifFalse: [interpreterProxy pop: 2 thenPush: interpreterProxy falseObject]
! !

!Win32OSProcessPlugin methodsFor: 'primitives - handles' stamp: 'dtl 3/29/2002 12:04'!
primitiveCloseHandle
	"Close the specified handle, which may refer to a process, a thread, or
	some other Win32 object."

	| handleOop handle result |
	self export: true.
	self var: 'handle' declareC: 'HANDLE handle'.
	self var: 'result' declareC: 'BOOL result'.
	handleOop _ interpreterProxy stackObjectValue: 0.
	(handle _ self handleFrom: handleOop) isNil ifTrue: [^ nil].
	result _ self CloseHandle: handle.
	interpreterProxy pop: 2.
	(result ~= 0)
		ifTrue: [interpreterProxy push: interpreterProxy trueObject]
		ifFalse: [interpreterProxy push: interpreterProxy falseObject]
! !

!Win32OSProcessPlugin methodsFor: 'primitives - console' stamp: 'dtl 3/25/2002 06:21'!
primitiveAllocConsole
	"Allocate a console if not already allocated."

	self export: true.
	self AllocConsole
! !

!Win32OSProcessPlugin methodsFor: 'primitives - console' stamp: 'dtl 3/25/2002 06:20'!
primitiveFreeConsole
	"Deallocate the console if allocated."

	self export: true.
	self FreeConsole
! !

!Win32OSProcessPlugin methodsFor: 'primitives - child process management' stamp: 'dtl 9/10/2002 09:17'!
primitiveCanAccessChildProcess
	"Answer true if the OS process represented by a HANDLE can be accessed by this OS process."

	| handle exitStatus |
	self export: true.
	self var: 'handle' declareC: 'HANDLE handle'.
	self var: 'exitStatus' declareC: 'DWORD exitStatus'.
	(handle _ self handleFrom: (interpreterProxy stackObjectValue: 0)) isNil ifTrue: [^ nil].
	(((self cCode: 'GetExitCodeProcess(handle, &exitStatus)' inSmalltalk: [0]) ~= 0)
		and: [self cCode: 'exitStatus == STILL_ACTIVE' inSmalltalk: [false]])
			ifTrue: [interpreterProxy pop: 2 thenPush: interpreterProxy trueObject]
			ifFalse: [interpreterProxy pop: 2 thenPush: interpreterProxy falseObject]! !

!Win32OSProcessPlugin methodsFor: 'primitives - child process management' stamp: 'dtl 3/29/2002 12:04'!
primitiveGetExitStatusForHandle
	"Answer the exit status for the process represented by a HANDLE. Fail if
	the process is still active, or if the GetExitCodeProcess call fails."

	| handle exitStatus |
	self export: true.
	self var: 'handle' declareC: 'HANDLE handle'.
	self var: 'exitStatus' declareC: 'DWORD exitStatus'.
	(handle _ self handleFrom: (interpreterProxy stackObjectValue: 0)) isNil ifTrue: [^ nil].
	((self cCode: 'GetExitCodeProcess(handle, &exitStatus)' inSmalltalk: [0]) ~= 0)
		ifTrue:
			[(self cCode: 'exitStatus == STILL_ACTIVE' inSmalltalk: [false]) ifTrue:
				["Child process is still active, fail the primitive"
				^ interpreterProxy primitiveFail].
			interpreterProxy
				pop: 2
				thenPush: (interpreterProxy positive32BitIntegerFor: exitStatus)]
		ifFalse:
			[^ interpreterProxy primitiveFail]
! !

!Win32OSProcessPlugin methodsFor: 'primitives - child process management' stamp: 'ajh 11/26/2002 20:50'!
primitiveOneShot
	"Anwer true the first time this primitive is called in a Squeak session, and false thereafter."

	| thisPrimHasBeenCalled |
	self export: true.
	self var: 'thisPrimHasBeenCalled' declareC: 'static int thisPrimHasBeenCalled= 0'.
	(thisPrimHasBeenCalled = 0)
		ifTrue:
			[thisPrimHasBeenCalled _ 1.
			interpreterProxy pop: 1 thenPush: interpreterProxy trueObject]
		ifFalse:
			[interpreterProxy pop: 1 thenPush: interpreterProxy falseObject]
! !

!Win32OSProcessPlugin methodsFor: 'primitives - child process management' stamp: 'dtl 2/27/2002 09:50'!
primitiveSetWaitForAnyProcessExitThenSignalSemaphoreWithIndex
	"Set up a thread to wait for a process HANDLE to exit, then signal the Semaphore
	at index. This provides asychronous notification of an external process exit. Answer
	the handle of the handler thread."

	| handleArrayOop handleArrayOopPtr handleArray handleArrayPtr handleCount arraySize idx semaIndex threadInformation |
	self export: true.
	self var: 'handleArrayOopPtr' type: 'unsigned int *'.
	self var: 'handleArray' declareC: 'static HANDLE handleArray[MAXIMUM_WAIT_OBJECTS]'.
	self var: 'handleArrayPtr' declareC: 'static HANDLE *handleArrayPtr'.
	self var: 'exitStatus' declareC: 'DWORD exitStatus'.
	semaIndex _ interpreterProxy stackIntegerValue: 0.
	handleArrayOop _ interpreterProxy stackObjectValue: 1.
	handleArrayOopPtr _ interpreterProxy firstIndexableField: handleArrayOop.
	handleCount _ interpreterProxy stSizeOf: handleArrayOop.
	arraySize _ self cCode: 'MAXIMUM_WAIT_OBJECTS' inSmalltalk: [64].
	(handleCount > arraySize) ifTrue: [^ interpreterProxy primitiveFail].
	idx _ 0.
	[idx < handleCount] whileTrue:
		[handleArray at: idx put: (self handleFrom: (handleArrayOopPtr at: idx)).
		idx _ idx + 1].
	handleArrayPtr _ handleArray.
	threadInformation _ self startHandlerThreadForAny: handleArrayPtr
							count: handleCount
							signalingSemaphoreAt: semaIndex.
	interpreterProxy pop: 3 thenPush: threadInformation
! !

!Win32OSProcessPlugin methodsFor: 'primitives - child process management' stamp: 'dtl 3/29/2002 12:05'!
primitiveTerminateThread
	"Kill the thread. No cleanup is performed, so use with caution for a thread which
	is (for example) manipulating a mutex. Answer true for success, else false."

	| handle |
	self export: true.
	self var: 'handle' declareC: 'HANDLE handle'.
	self var: 'exitStatus' declareC: 'DWORD exitStatus'.
	(handle _ self handleFrom: (interpreterProxy stackObjectValue: 0)) isNil ifTrue: [^ nil].
	(self cCode: 'TerminateThread(handle, 0)' inSmalltalk: [false])
		ifTrue: [^ interpreterProxy pop: 2 thenPush: interpreterProxy trueObject]
		ifFalse: [^ interpreterProxy pop: 2 thenPush: interpreterProxy falseObject]
! !

!Win32OSProcessPlugin methodsFor: 'primitives - nonblocking read' stamp: 'dtl 8/17/2002 12:49'!
primitiveBufferValuesAt
	"For debugging only. Answer the current values of readCharBufferArray,
	readCharCountArray, and readCharStatusArray at index, an integer corresponding
	to a semaphore for one read handler thread. Answer an Array with the buffered
	character, the character count, and the status value."

	| index result byte count status |
	self export: true.
	self var: 'byte' declareC: 'unsigned char byte'.
	self var: 'count' declareC: 'DWORD count'.
	self var: 'status' declareC: 'int status'.
	index _ interpreterProxy stackIntegerValue: 0.
	byte _ readCharBufferArray at: index.
	count _ readCharCountArray at: index.
	status _ readCharStatusArray at: index.
	result _ interpreterProxy
		instantiateClass: interpreterProxy classArray
		indexableSize: 3.
	interpreterProxy
		storePointer: 0
		ofObject: result
		withValue: (interpreterProxy positive32BitIntegerFor: byte).
	interpreterProxy
		storePointer: 1
		ofObject: result
		withValue: (interpreterProxy positive32BitIntegerFor: count).
	interpreterProxy
		storePointer: 2
		ofObject: result
		withValue: (interpreterProxy positive32BitIntegerFor: status).
	interpreterProxy pop: 2 thenPush: result
! !

!Win32OSProcessPlugin methodsFor: 'primitives - nonblocking read' stamp: 'dtl 4/1/2002 19:39'!
primitiveLastReadFor
	"A character has been read into an external buffer corresponding to aSemaphoreIndex,
	and is now available. Answer the integer value of the character, or nil if no character
	was read, or -1 if an error occurred on the read."

	| index result |
	self export: true.
	index _ interpreterProxy stackIntegerValue: 0.
	(readCharStatusArray at: index)
		ifFalse:
			[result _ interpreterProxy integerObjectOf: -1]
		ifTrue:
			[((readCharCountArray at: index) < 1)
				ifTrue:
					[result _ interpreterProxy nilObject]
				ifFalse:
					[result _ interpreterProxy integerObjectOf: (readCharBufferArray at: index)]].
	interpreterProxy pop: 2 thenPush: result
! !

!Win32OSProcessPlugin methodsFor: 'primitives - nonblocking read' stamp: 'dtl 8/28/2002 07:42'!
primitiveLastReadForStoreIn
	"A character has been read into an external buffer corresponding to aSemaphoreIndex,
	and is now available. Answer the integer value of the character, or nil if no character
	was read, or -1 if an error occurred on the read. Store the full results of the most
	recent read into a three element array provided by the sender. Contents of the array
	will be status of the read call, character read, and character count (which should
	always be 1)."

	| index resultArray result |
	self export: true.
	index _ interpreterProxy stackIntegerValue: 1.
	resultArray _ interpreterProxy stackValue: 0.
	interpreterProxy stObject: resultArray at: 1 put:
		(interpreterProxy positive32BitIntegerFor: (readCharStatusArray at: index)).
	interpreterProxy stObject: resultArray at: 2 put:
		(interpreterProxy integerObjectOf: (readCharBufferArray at: index)).
	interpreterProxy stObject: resultArray at: 3 put:
		(interpreterProxy integerObjectOf: (readCharCountArray at: index)).
	(readCharStatusArray at: index)
		ifFalse:
			[result _ interpreterProxy integerObjectOf: -1]
		ifTrue:
			[((readCharCountArray at: index) < 1)
				ifTrue:
					[result _ interpreterProxy nilObject]
				ifFalse:
					[result _ interpreterProxy integerObjectOf: (readCharBufferArray at: index)]].
	interpreterProxy pop: 3 thenPush: result
! !

!Win32OSProcessPlugin methodsFor: 'primitives - nonblocking read' stamp: 'dtl 9/18/2002 09:39'!
primitiveNextFromSignaling
	"Read the next character from anIOHandle (a SQFile struct) into a buffer in the VM. When
	the read completes, signal the specified Semaphore to notify that the character is available.
	Answer the handle of the new thread, or nil on error."

	| index handle hThread threadArgs threadID |
	self export: true.
	self var: 'index' type: 'static int'.
	self var: 'handle' type: 'HANDLE'.
	self var: 'hThread' type: 'HANDLE'.
	self var: 'threadID' declareC: 'DWORD threadID'.
	self var: 'threadArgs' declareC: 'static struct {int index; HANDLE handle;} threadArgs'.

	index _ interpreterProxy stackIntegerValue: 0.
	(index >= readCharArraySize) ifTrue: [^ interpreterProxy primitiveFail].

	"Complete setup of one handler before doing the next. Mutex will be
	released in the newly started thread."
	self waitForThreadMutex: self readThreadMutexHandle.

	handle _ self handleFromSQFile: (interpreterProxy stackObjectValue: 1).
	handle isNil ifTrue: [^ nil]. "Fail if not a handle, probably due to file having been closed"
	readCharBufferArray at: index put: 0.
	readCharCountArray at: index put: 0.
	readCharStatusArray at: index put: 0.
	self cCode: 'threadArgs.index = index'.
	self cCode: 'threadArgs.handle = handle'.
	hThread _ self cCode: 'CreateThread(NULL, 1024, readCharThread, &threadArgs, 0, &threadID)'.
	"Safety net: Remember the identity of the thread which should be active for this
	semaphore index."
	readCharThreadIdArray at: index put: threadID.
	hThread isNil ifTrue: [^ interpreterProxy nilObject].
	interpreterProxy pop: 3 thenPush: (self handleObjectFrom: hThread)
! !

!Win32OSProcessPlugin methodsFor: 'private - thread synchronization' stamp: 'dtl 4/6/2002 19:45'!
createThreadMutex
	"Answer a HANDLE for a new mutex semaphore."

	self returnTypeC: 'HANDLE'.
	^ self cCode: 'CreateMutex(NULL, false, NULL)'
! !

!Win32OSProcessPlugin methodsFor: 'private - thread synchronization' stamp: 'dtl 8/24/2002 16:29'!
exitThreadMutexHandle
	"Answer a HANDLE for a mutex semaphore for the child handler thread."

	| h |
	self returnTypeC: 'HANDLE'.
	self var: 'h' declareC: 'static HANDLE h= 0'.
	h isNil ifTrue: [h _ self createThreadMutex].
	^ h
! !

!Win32OSProcessPlugin methodsFor: 'private - thread synchronization' stamp: 'dtl 8/25/2002 11:07'!
readThreadMutexHandle
	"Answer a HANDLE for a mutex semaphore for sychronizing setup of read threads."

	| h |
	self returnTypeC: 'HANDLE'.
	self var: 'h' declareC: 'static HANDLE h= 0'.
	h isNil ifTrue: [h _ self createThreadMutex].
	^ h
! !

!Win32OSProcessPlugin methodsFor: 'private - thread synchronization' stamp: 'dtl 8/24/2002 10:56'!
releaseThreadMutex: aHandle
	"To be called at the end of a critical section."

	self returnTypeC: 'BOOL'.
	self var: 'aHandle' type: 'HANDLE'.
	^ self ReleaseMutex: aHandle
! !

!Win32OSProcessPlugin methodsFor: 'private - thread synchronization' stamp: 'dtl 8/25/2002 13:22'!
signalSemaphoreMutexHandle
	"Answer a HANDLE for a mutex semaphore for event notification."

	| h |
	self returnTypeC: 'HANDLE'.
	self var: 'h' declareC: 'static HANDLE h= 0'.
	h isNil ifTrue: [h _ self createThreadMutex].
	^ h
! !

!Win32OSProcessPlugin methodsFor: 'private - thread synchronization' stamp: 'dtl 8/25/2002 13:25'!
threadSafeSignalSemaphoreWithIndex: index

	"This method should call synchronizedSignalSemaphoreWithIndex() in
	sqWin32Window.c in order to ensure that multiple threads do not execute
	vm code concurrently. Unfortunately (FIXME) I do not know how to call
	synchronizedSignalSemaphoreWithIndex() from a DLL, so this will have to
	do for now."

	self waitForThreadMutex: self signalSemaphoreMutexHandle.
	interpreterProxy signalSemaphoreWithIndex: index.
	self releaseThreadMutex: self signalSemaphoreMutexHandle
! !

!Win32OSProcessPlugin methodsFor: 'private - thread synchronization' stamp: 'dtl 8/24/2002 10:56'!
waitForThreadMutex: aHandle
	"To be called before entering a critical section."

	self returnTypeC: 'DWORD'.
	self var: 'aHandle' type: 'HANDLE'.
	^ self cCode: 'WaitForSingleObject(aHandle, INFINITE)'
! !

!Win32OSProcessPlugin methodsFor: 'private - pipes' stamp: 'dtl 3/11/2005 18:25'!
createPipeForReader: readerIOStreamPtr writer: writerIOStreamPtr
	"Create a pipe and populate the readerIOStream and writerIOStream variables.
	Answer true for success, else false"

	self var: 'readerIOStreamPtr' type: 'FILEHANDLETYPE *'.
	self var: 'writerIOStreamPtr' type: 'FILEHANDLETYPE *'.
	^ self cCode: 'CreatePipe(readerIOStreamPtr, writerIOStreamPtr, NULL, 0)'
! !

!Win32OSProcessPlugin methodsFor: 'private - file handles' stamp: 'dtl 3/29/2002 14:33'!
stderrHandle

	| h |
	self returnTypeC: 'HANDLE'.
	self var: 'h' type: 'HANDLE'.
	h _ self cCode: 'GetStdHandle(STD_ERROR_HANDLE)' inSmalltalk: [0].
	(self cCode: 'h == INVALID_HANDLE_VALUE' inSmalltalk: [-1])
		ifTrue: [interpreterProxy primitiveFail].
	^ h
! !

!Win32OSProcessPlugin methodsFor: 'private - file handles' stamp: 'dtl 3/29/2002 14:33'!
stdinHandle

	| h |
	self returnTypeC: 'HANDLE'.
	self var: 'h' type: 'HANDLE'.
	h _ self cCode: 'GetStdHandle(STD_INPUT_HANDLE)' inSmalltalk: [0].
	(self cCode: 'h == INVALID_HANDLE_VALUE' inSmalltalk: [-1])
		ifTrue: [interpreterProxy primitiveFail].
	^ h
! !

!Win32OSProcessPlugin methodsFor: 'private - file handles' stamp: 'dtl 3/29/2002 14:33'!
stdoutHandle

	| h |
	self returnTypeC: 'HANDLE'.
	self var: 'h' type: 'HANDLE'.
	h _ self cCode: 'GetStdHandle(STD_OUTPUT_HANDLE)' inSmalltalk: [0].
	(self cCode: 'h == INVALID_HANDLE_VALUE' inSmalltalk: [-1])
		ifTrue: [interpreterProxy primitiveFail].
	^ h
! !

!Win32OSProcessPlugin methodsFor: 'private - exit handler thread' stamp: 'dtl 8/24/2002 19:09'!
handleAnyChildProc: args
	"Wait for a process HANDLE to exit, then signal the Semaphore at index. Answer
	the exit status of the process. This is the code to be executed in a thread to provide
	asychronous notification of an external process exit."

	| threadArgs handles count index exitStatus |
	self returnTypeC: 'DWORD WINAPI'.
	self var: 'args' type: 'void *'.
	self var: 'threadArgs' type: 'struct {int count; HANDLE *handles; int semaIndex;} *'.
	self var: 'handles' declareC: 'volatile LPHANDLE handles'.
	self var: 'count' declareC: 'volatile int count'.
	self var: 'index' declareC: 'volatile int index'.
	self var: 'exitStatus' declareC: 'DWORD exitStatus'.
	threadArgs _ args.
	self cCode: 'count= threadArgs->count' inSmalltalk: [0].
	self cCode: 'handles= threadArgs->handles' inSmalltalk: [0].
	self cCode: 'index= threadArgs->semaIndex' inSmalltalk: [0].

	"Release mutex, allowing additional threads to be set up"
	self releaseThreadMutex: self exitThreadMutexHandle.

	exitStatus _ self cCode: 'WaitForMultipleObjects(count, handles, FALSE, INFINITE)'.
	self threadSafeSignalSemaphoreWithIndex: index.
	^ exitStatus
! !

!Win32OSProcessPlugin methodsFor: 'private - exit handler thread' stamp: 'dtl 8/24/2002 16:31'!
startHandlerThreadForAny: handleArrayPointer count: arraySize signalingSemaphoreAt: semaphoreIndex
	"Wait for a process HANDLE to exit, then signal the Semaphore at index. Answer
	the thread ID of the handler. This is the code to be executed in a thread to provide
	asychronous notification of an external process exit."

	| threadID hThread threadInformation threadArgs |
	self var: 'handleArrayPointer' type: 'HANDLE *'.
	self var: 'threadID' declareC: 'DWORD threadID'.
	self var: 'hThread' declareC: 'HANDLE hThread'.
	self var: 'threadArgs' declareC: 'static struct {int count; HANDLE *handles; int semaIndex;} threadArgs'.

	"Complete setup of one handler before doing the next. Mutex will be
	released in the newly started thread."
	self waitForThreadMutex: self exitThreadMutexHandle.

	"Set up the thread argument structure."
	self cCode: 'threadArgs.count= arraySize'.
	self cCode: 'threadArgs.handles= handleArrayPointer'.
	self cCode: 'threadArgs.semaIndex= semaphoreIndex'.

	hThread _ self cCode: 'CreateThread(NULL, 1024, handleAnyChildProc, &threadArgs, 0, &threadID)'.
	hThread isNil ifTrue: [^ interpreterProxy nilObject].
	interpreterProxy pushRemappableOop: (interpreterProxy positive32BitIntegerFor: threadID).
	interpreterProxy pushRemappableOop: (self handleObjectFrom: hThread).
	threadInformation _ interpreterProxy
		instantiateClass: interpreterProxy classArray
		indexableSize: 2.
	interpreterProxy
		storePointer: 0
		ofObject: threadInformation
		withValue: interpreterProxy popRemappableOop.
	interpreterProxy
		storePointer: 1
		ofObject: threadInformation
		withValue: interpreterProxy popRemappableOop.
	^ threadInformation
! !

!Win32OSProcessPlugin methodsFor: 'private - exit handler thread' stamp: 'dtl 2/26/2002 14:17'!
threadMutexHandle
	"Answer a HANDLE for a mutex semaphore."

	| h |
	self returnTypeC: 'HANDLE'.
	self var: 'h' declareC: 'static HANDLE h= 0'.
	(h = 0) ifTrue: [h _ self cCode: 'CreateMutex(NULL, false, NULL)'].
	^ h
! !

!Win32OSProcessPlugin methodsFor: 'private - nonblocking read thread' stamp: 'dtl 9/1/2002 09:21'!
readCharThreadIndex: index handle: handle
	"This is the thread procedure for reading one character from a file or pipe
	handle. The values of index and handle have been passed on the local
	thread stack, so they are now local to the thread which is executing this
	method. Once the values are safely on a local thread stack, the mutex
	semaphore is released, allowing other threads to be created."

	| status |
	self returnTypeC: 'DWORD WINAPI'.
	self var: 'handle' type: 'HANDLE'.
	"Release mutex, allowing additional threads to be set up"
	self releaseThreadMutex: self readThreadMutexHandle.
	"Read a character, possibly waiting indefinitely"
	status _ self cCode: 'ReadFile(handle, &readCharBufferArray[index], 1, &readCharCountArray[index], NULL)'.
	readCharStatusArray at: index put: status.
	"Ensure that this thread is still valid for this semaphore index."
	self GetCurrentThreadId = (readCharThreadIdArray at: index)
		ifTrue:
			["Signal Squeak that the character is now available"
			self threadSafeSignalSemaphoreWithIndex: index]
		ifFalse:
			["Invalidate any data we just placed in the buffer"
			readCharBufferArray at: index put: 0.
			readCharCountArray at: index put: 0.
			readCharStatusArray at: index put: 0].
	^ status
! !

!Win32OSProcessPlugin methodsFor: 'private - nonblocking read thread' stamp: 'dtl 8/24/2002 10:09'!
readCharThread: args
	"This is the entry point for the thread procedure for reading one character from
	a file or pipe handle. There may be many instances of this thread procedure, in
	which case each thread may be distinguished by the index value with which it is
	called. Remember the index value by saving it in a variable on the thread stack.
	Read one character, and signal a Smalltalk Semaphore to indicate completion."

	| index handle threadArgs |
	self returnTypeC: 'DWORD WINAPI'.
	self var: 'handle' type: 'HANDLE'.
	self var: 'args' type: 'void *'.
	self var: 'threadArgs' type: 'struct {int index; HANDLE handle;} *'.
	threadArgs _ args.
	self cCode: 'index = threadArgs->index'.
	self cCode: 'handle = threadArgs->handle'.
	"Save arguments in stack local variables. The mutex semaphore will be
	released as soon as these variables are safely on a local thread stack."
	^ self readCharThreadIndex: index handle: handle
! !

!Win32OSProcessPlugin methodsFor: 'private - handle conversion' stamp: 'dtl 9/18/2002 09:37'!
handleFromSQFile: anSQFileRecord

	| sqFile |
	self returnTypeC: 'HANDLE'.
	self var: 'sqFile' type: 'struct SQFile'.
	((interpreterProxy isBytes: anSQFileRecord) and:
			[(interpreterProxy stSizeOf: anSQFileRecord) = self fileRecordSize])
		ifFalse:
			[interpreterProxy primitiveFail.
			^ nil].
	sqFile _ self fileValueOf: anSQFileRecord.
	^ self cCode: 'sqFile->file' inSmalltalk: [0]
! !

!Win32OSProcessPlugin methodsFor: 'private - handle conversion' stamp: 'dtl 3/29/2002 11:40'!
handleFrom: aHandleObject
	"Answer the HANDLE represented by aHandleObject."

	| ptr handleUnion idx |
	self returnTypeC: 'HANDLE'.
	self var: 'ptr' declareC: 'unsigned char *ptr'.
	self var: 'handleUnion'
		declareC: 'union {void *address; unsigned char bytes[sizeof(void *)];} handleUnion'.
	((interpreterProxy isBytes: aHandleObject) and:
			[(interpreterProxy stSizeOf: aHandleObject) = self sizeOfHandle])
		ifFalse:
			[interpreterProxy primitiveFail.
			^ nil].
	ptr _ interpreterProxy arrayValueOf: aHandleObject.
	idx _ 0.
	[idx < self sizeOfHandle]	whileTrue:
		[self cCode: 'handleUnion.bytes[idx] = ptr[idx]'.
		idx _ idx + 1].
	^ self cCode: 'handleUnion.address' inSmalltalk: [handleUnion]
! !

!Win32OSProcessPlugin methodsFor: 'private - handle conversion' stamp: 'dtl 2/23/2002 13:53'!
handleObjectFrom: aHandle
	"Answer an object which contains the value of a HANDLE."
 
	| size handleOop ptr handleUnion idx |
	self var: 'aHandle' type: 'HANDLE'.
	self var: 'ptr' declareC: 'unsigned char *ptr'.
	self var: 'handleUnion' declareC: 'union {HANDLE handle; char bytes[sizeof(HANDLE)];} handleUnion'.
	self cCode: 'handleUnion.handle = aHandle'.
	size _ self sizeOfHandle.
	handleOop _ interpreterProxy
		instantiateClass: interpreterProxy classByteArray
		indexableSize: size.
	ptr _ interpreterProxy arrayValueOf: handleOop.
	idx _ 0.
	[idx < size] whileTrue:
		[self cCode: 'ptr[idx]= handleUnion.bytes[idx]'.
		idx _ idx + 1].
	^ handleOop
! !

!Win32OSProcessPlugin methodsFor: 'private - handle conversion' stamp: 'dtl 2/23/2002 13:35'!
sizeOfHandle

	^ self cCode: 'sizeof(HANDLE)' inSmalltalk: [4]
! !

!Win32OSProcessPlugin methodsFor: 'private - handle conversion' stamp: 'dtl 9/8/2002 22:23'!
stdErrHandle
	"Answer the pseudo HANDLE for standard error."

	self returnTypeC: 'HANDLE'.
	^ self cCode: '(HANDLE) STD_ERROR_HANDLE' inSmalltalk: ['16rFFFFFFF4' asNumber]
! !

!Win32OSProcessPlugin methodsFor: 'private - handle conversion' stamp: 'dtl 9/8/2002 22:23'!
stdInHandle
	"Answer the pseudo HANDLE for standard input."

	self returnTypeC: 'HANDLE'.
	^ self cCode: '(HANDLE) STD_INPUT_HANDLE' inSmalltalk: ['16rFFFFFFF6' asNumber]
! !

!Win32OSProcessPlugin methodsFor: 'private - handle conversion' stamp: 'dtl 9/8/2002 22:23'!
stdOutHandle
	"Answer the pseudo HANDLE for standard output."

	self returnTypeC: 'HANDLE'.
	^ self cCode: '(HANDLE) STD_OUTPUT_HANDLE' inSmalltalk: ['16rFFFFFFF5' asNumber]
! !

!Win32OSProcessPlugin methodsFor: 'deprecated' stamp: 'dtl 3/3/2002 07:16'!
makePipeForReader: readerIOStreamPtr writer: writerIOStreamPtr
	"Create a pipe and populate the readerIOStream and writerIOStream variables.
	Answer true for success, else false"

	self var: 'readerIOStreamPtr' type: 'FILEHANDLETYPE *'.
	self var: 'writerIOStreamPtr' type: 'FILEHANDLETYPE *'.
	^ self cCode: 'CreatePipe(readerIOStreamPtr, writerIOStreamPtr, NULL, 0)'
! !

!Win32OSProcessPlugin methodsFor: 'deprecated' stamp: 'dtl 9/9/2002 09:23'!
primitiveCommand
	"Create a process to run commandString. Answer an Array formed from the
	PROCESS_INFORMATION structure for the new process. The caller is expected
	to close the process and thread handles when they are no longer needed.
	The result array contains hProcess, hThread, dwProcessId, dwThreadId.

	This primitive has been replaced by #primitiveCommandWithInputOutputError,
	and will be removed in future versions of OSProcess."

	| cStringPtr commandString startUp procInfo processInformation |
	self export: true.
	self var: 'cStringPtr' declareC: 'LPTSTR cStringPtr'.
	self var: 'startUp' declareC: 'STARTUPINFO startUp'.
	self var: 'procInfo' declareC: 'PROCESS_INFORMATION procInfo'.

	commandString _ interpreterProxy stackObjectValue: 0.
	cStringPtr _ self cStringFromString: commandString.

	self cCode: 'GetStartupInfo (&startUp)'.
	self cCode: 'startUp.hStdInput = GetStdHandle(STD_INPUT_HANDLE)'.
	self cCode: 'startUp.hStdOutput = GetStdHandle(STD_OUTPUT_HANDLE)'.
	self cCode: 'startUp.hStdError = GetStdHandle(STD_ERROR_HANDLE)'.
	(self cCode:  'CreateProcess(NULL, cStringPtr, NULL, NULL, TRUE, NORMAL_PRIORITY_CLASS, NULL, NULL, &startUp, &procInfo)')
		ifFalse: [^ interpreterProxy primitiveFail].

	interpreterProxy pushRemappableOop: (interpreterProxy positive32BitIntegerFor: (self cCode: 'procInfo.dwThreadId')).
	interpreterProxy pushRemappableOop: (interpreterProxy positive32BitIntegerFor: (self cCode: 'procInfo.dwProcessId')).
	interpreterProxy pushRemappableOop: (self handleObjectFrom: (self cCode: 'procInfo.hThread')).
	interpreterProxy pushRemappableOop: (self handleObjectFrom: (self cCode: 'procInfo.hProcess')).
	processInformation _ interpreterProxy
		instantiateClass: interpreterProxy classArray
		indexableSize: 4.
	interpreterProxy
		storePointer: 0
		ofObject: processInformation
		withValue: interpreterProxy popRemappableOop.
	interpreterProxy
		storePointer: 1
		ofObject: processInformation
		withValue: interpreterProxy popRemappableOop.
	interpreterProxy
		storePointer: 2
		ofObject: processInformation
		withValue: interpreterProxy popRemappableOop.
	interpreterProxy
		storePointer: 3
		ofObject: processInformation
		withValue: interpreterProxy popRemappableOop.
	interpreterProxy pop: 2; push: processInformation
! !


!Win32OSProcessPlugin class methodsFor: 'translation' stamp: 'dtl 9/4/2002 19:47'!
declareCVarsIn: cg

	"Additional declarations are done in my subclasses. This permits a meaningful
	module identification string to be written with the name of the actual subclass
	being generated."

	cg addHeaderFile: '<windows.h>'.
	cg addHeaderFile: '<process.h>'.
	cg addHeaderFile: '"sqWin32.h"
#define FILEHANDLETYPE HANDLE	/* The type of low level stream to be used in a struct SQFile */
#define READCHARARRAYSIZE 512	/* Corresponds to expected max size of external semaphore table */
									/* Table expands dynamically, so just use a big number */'.
	cg var: 'readCharArraySize' declareC: 'int readCharArraySize = READCHARARRAYSIZE'.
	cg var: 'readCharBufferArray' declareC: 'unsigned char readCharBufferArray[READCHARARRAYSIZE]'.
	cg var: 'readCharCountArray' declareC: 'DWORD readCharCountArray[READCHARARRAYSIZE]'.
	cg var: 'readCharStatusArray' declareC: 'int readCharStatusArray[READCHARARRAYSIZE]'.
	cg var: 'readCharThreadIdArray' declareC: 'DWORD readCharThreadIdArray[READCHARARRAYSIZE]'.
! !

!Win32OSProcessPlugin class methodsFor: 'translation' stamp: 'dtl 2/21/2002 16:45'!
moduleName

	^ 'Win32OSProcessPlugin'! !

!Win32OSProcessPlugin class methodsFor: 'translation' stamp: 'dtl 3/3/2002 13:07'!
shouldBeTranslated

	^ self isResponsibleForThisPlatform and: [self concreteClass == self]
! !

!Win32OSProcessPlugin class methodsFor: 'concrete subclasses' stamp: 'dtl 8/7/2005 22:46'!
concreteClass
	"Win32OSProcessPlugin concreteClass"

	^ self hasInterpreterGetThisSession
		ifTrue: [Win32OSProcessPluginInterpreterGetThisSession]
		ifFalse: [Win32OSProcessPluginNoThisSessionAvailable]
! !

!Win32OSProcessPlugin class methodsFor: 'concrete subclasses' stamp: 'dtl 3/5/2005 12:11'!
isResponsibleForThisPlatform
	"Answer true is this is an instance of the class which is responsible for representing
	the OS process for the Squeak VM running on the current platform. A false answer is
	usually the result of running the image on a different platform and VM."

	^ self platformName = 'Win32'! !

!Win32OSProcessPlugin class methodsFor: 'version testing' stamp: 'dtl 3/1/2002 17:06'!
installedModuleName
	"Answer the module name string for the plugin which is actually running in the VM"

	"Win32OSProcessPlugin installedModuleName"

	<primitive: 'primitiveModuleName' module: 'Win32OSProcessPlugin'>
	^ nil
! !

!Win32OSProcessPlugin class methodsFor: 'version testing' stamp: 'dtl 3/1/2002 17:24'!
installedModuleVersionString
	"Answer the version string for the plugin which is actually running in the VM"

	"Win32OSProcessPlugin installedModuleVersionString"

	<primitive: 'primitiveVersionString' module: 'Win32OSProcessPlugin'>
	^ nil
! !

!Win32OSProcessPlugin class methodsFor: 'version testing' stamp: 'dtl 8/13/2003 21:55'!
primGetPid
	"Answer the OS process ID for the OS process in which I am currently executing."

	<primitive: 'primitiveGetPid' module: 'Win32OSProcessPlugin'>
	^ nil! !


!Win32OSProcessPluginInterpreterGetThisSession methodsFor: 'private' stamp: 'dtl 8/8/2005 06:25'!
getThisSessionIdentifier

	^ interpreterProxy getThisSessionID
! !


!Win32OSProcessPluginInterpreterGetThisSession class methodsFor: 'translation' stamp: 'dtl 8/8/2005 06:20'!
declareCVarsIn: cg

	cg addHeaderFile: '<sys/types.h>
/* ', self idString, ' - ', self moduleName, '.c translated from class
   ', self name, ' of OSProcessPlugin version ', self versionString, ' */'
! !


!Win32OSProcessPluginNoThisSessionAvailable methodsFor: 'private' stamp: 'dtl 3/1/2002 13:25'!
getThisSessionIdentifier

	^ nil! !


!Win32OSProcessPluginNoThisSessionAvailable class methodsFor: 'translation' stamp: 'dtl 8/8/2005 06:20'!
declareCVarsIn: cg

	cg addHeaderFile: '<sys/types.h>
/* ', self idString, ' - ', self moduleName, '.c translated from class
   ', self name, ' of OSProcessPlugin version ', self versionString, ' */'
! !

